# nxpy.svn package -----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2016
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Subversion client wrapper.

Only supports versions 1.6, 1.7 and 1.8, others might work but have not been
tested.
Requires at least Python 2.6.

"""

# Ensures that tests are skipped rather than broken when run with Python 2.5
from __future__ import with_statement
from __future__ import absolute_import
 
import logging
import os
import os.path
import re

import six

import nxpy.command.command
import nxpy.command.option
import nxpy.core.error
import nxpy.core.past
import nxpy.core.temp_file
import nxpy.svn.url

import nxpy.core._impl.log

_log = logging.getLogger(__name__)


class Info(object):
    r"""Represents the output of the ``svn info`` command in a structured way."""
    _outRe = re.compile(r"([^:]+):\s+(.*)")
    
    def __init__(self, out):
        for l in out.split("\n"):
            match = Info._outRe.match(l)
            if match:
                if match.group(1) == "Path":
                    self.path = os.path.realpath(match.group(2))
                elif match.group(1) == "URL":
                    self.url = match.group(2)
                elif match.group(1) == "Repository Root":
                    self.root = match.group(2)
    
    def __str__(self):
        return " - ".join((self.path, str(self.url), self.root))


class Status(object):
    r"""Represents the output of one line of the ``svn status`` command in a structured way."""
    
    def __init__(self, line):
        self.item = line[8:]
        self.state = line[0]
        self.properties = line[1]
        self.locked = line[2] == 'L'
        self.add_with_history = line[3] == '+'
        self.switched = line[4] == 'S'
        self.lock = line[5]
        self.tree_conflict = line[6] == 'C'

    def __str__(self):
        return ( self.state + self.properties + ( 'L' if self.locked else ' ') + 
                ( '+' if self.add_with_history else ' ' ) + ( 'S' if self.switched else ' ' ) +
                self.lock + ( 'C' if self.tree_conflict else ' ' ) + ' ' + self.item )


_config = nxpy.command.option.Config(
        bool_opts = ( "ignore_externals", "keep_local", "parents", "quiet", "stop_on_copy", 
                "summarize", "verbose", "version", "xml" ),
        value_opts = ( "file", "message", "password", "revision", "username", "value" ),
        opposite_opts = { "ignore" : "--no-ignore" },
        mapped_opts = { "ignore_externals" : "--ignore-externals",
                        "keep_local" : "--keep-local",
                        "stop_on_copy" : "--stop-on-copy",
                        "value" : "",
                        "version" : "--version" } )


class Parser(nxpy.command.option.Parser):
    r"""Allows passing *nxpy.svn.url.Url* instances as arguments to *Svn*'s methods."""
    def __init__(self, command, arguments, options, **defaults):
        super(Parser, self).__init__(_config, command, 
                [ self._pathtostring(a) for a in arguments ], options, **defaults)
    
    def _pathtostring(self, path):
        if isinstance(path, nxpy.svn.url.Url):
            return str(path)
        elif isinstance(path, six.string_types):
            return path
        else:
            raise nxpy.core.error.ArgumentError()
        

class Svn(nxpy.command.command.Command):
    r"""The actual wrapper."""
    def __init__(self, debug=None):
        super(Svn, self).__init__("svn --non-interactive", debug)
        self._version = None
    
    _versionRe = re.compile(r"(\d+)\.(\d+)\.(\d+)")
    
    def version(self):
        op = Parser("", (), {}, version=True)
        out = self.run(op)[0]
        match = Svn._versionRe.search(out)
        version = ( int(match.group(1)), int(match.group(2)), int(match.group(3)))
        return version

    def info(self, *targets):
        op = Parser("info", targets, {})
        out = self.run(op)[0]
        return Info(out)

    def list(self, *targets):
        op = Parser("list", targets, {})
        out = self.run(op)[0]
        return out.split()
    
    def mkdir(self, *targets, **options):
        op = Parser("mkdir", targets, options, parents=False, username="", password="",
                message="\"[" + __name__ + "] Make dir(s) " + ", ".join(targets) + "\"")
        self.run(op)

    def status(self, *targets, **options):
        op = Parser("status", targets, options, ignore=True, quiet=True, ignore_externals=True)
        out = self.run(op)[0]
        ignore_externals = options.get('ignore_externals', True)
        res = []
        for l in out.split("\n"):
            if len(l) > 0 and ( not ignore_externals or l[0] != "X" ):
                res.append(Status(l))
        return res

    def import_(self, src, dest, debug=None, **options):
        op = Parser("import", ( src, dest ), {}, username="", password="", 
                    message="\"[" + __name__ + "] Import from " + src + "\"")
        self.run(op, debug)

    def checkout(self, src, dest, debug=None, **options):
        op = Parser("checkout", ( src, dest ), options, username="", password="", 
                ignore_externals=True)
        self.run(op, debug)

    def update(self, *targets, **options):
        op = Parser("update", targets, options, ignore_externals=True)
        self.run(op)

    def commit(self, src, debug=None, **options):
        op = Parser("commit", ( src, ), options, username="", password="", 
                    message="\"[" + __name__ + "] Commit from " + src + "\"")
        self.run(op, debug)

    def copy(self, src, dest, debug=None, **options):
        op = Parser("copy", ( src, dest ), options, ignore_externals=True, username="", password="", 
                    message="\"[" + __name__ + "] Copy from " + src + "\"")
        self.run(op, debug)

    def _delete_paths(self, debug, *paths, **options):
        op = Parser("delete", paths, options, keep_local=False)
        self.run(op, debug)

    def _delete_urls(self, debug, *urls, **options):
        op = Parser("delete", urls, options, username="", password="", 
                message="\"[" + __name__ + "] Delete " + ", ".join([ str(u) for u in urls ]) + "\"")
        self.run(op, debug)

    def delete(self, *targets, **options):
        debug = options.get("debug")
        if debug is not None:
            del options["debug"]
        paths = []
        urls = []
        for t in targets:
            if not isinstance(t, nxpy.svn.url.Url) and os.access(t, os.R_OK):
                paths.append(t)
            else:
                urls.append(t)
        if len(paths) > 0:
            self._delete_paths(debug, *paths, **options)
        if len(urls) > 0:
            self._delete_urls(debug, *urls, **options)
    
    def propget(self, name, *targets):
        op = Parser("propget", ( name, ) + targets, {})
        out = self.run(op)[0]
        return out.strip()

    _extRe = re.compile(r"([\S]+)\s+(.*)")
    
    def getexternals(self, d):
        out = self.propget("svn:externals", d)
        res = {}
        for l in out.split("\n"):
            match = Svn._extRe.match(l)
            if match:
                res[match.group(1)] = match.group(2)
        return res
    
    def getignore(self, d):
        out = self.propget("svn:ignore", d)
        return list(out.split("\n"))
    
    def propset(self, name, *targets, **options):
        debug = options.get("debug")
        value = options.get("value")
        if value:
            del options["value"]
            op = Parser("propset", ( name, value ) + targets, options, username="", password="")
        else:
            op = Parser("propset", ( name, ) + targets, options, file="", username="", password="")
        self.run(op, debug)
        
    def setexternals(self, externals, d, username="", password=""):
        nxpy.core.past.enforce_at_least(nxpy.core.past.V_2_6)
        with nxpy.core.temp_file.TempFile(mode="w+", prefix="svn_setexternals") as f:
            for k, v in six.iteritems(externals):
                f.write(k + "\t" + v + "\n")
            f.seek(0, os.SEEK_SET)
            _log.debug(f.read())
            f.close()
            self.propset("svn:externals", d, file=f.name, username=username, password=password)

    def setignore(self, ignore, d, username="", password=""):
        nxpy.core.past.enforce_at_least(nxpy.core.past.V_2_6)
        with nxpy.core.temp_file.TempFile(mode="w+", prefix="svn_setignore") as f:
            for v in ignore:
                f.write(v + "\n")
            f.seek(0, os.SEEK_SET)
            _log.debug(f.read())
            f.close()
            self.propset("svn:ignore", d, file=f.name, username=username, password=password)

    def diff(self, *targets, **options):
        op = Parser("diff", targets, options, summarize=False, revision=None)
        out = self.run(op)[0]
        return out

    def cat(self, *targets, **options):
        op = Parser("cat", targets, options, revision=None)
        out = self.run(op)[0]
        return out

    def log(self, src, **options):
        op = Parser("log", ( src, ), options, verbose=False, stop_on_copy=False, revision=None, 
                xml=False)
        out = self.run(op)[0]
        return out



# nxpy.msvs._impl.solution package -------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

from __future__ import absolute_import

import logging

import nxpy.ply.parser

import nxpy.core._impl.log
import nxpy.msvs._impl.solution.ast


_log = logging.getLogger(__name__)


class Parser(nxpy.ply.parser.Parser):
    def __init__(self, *args, **kwargs):
        super(Parser, self).__init__(*args, **kwargs)
        self.desc_pos = -1

    def p_solution(self, t):
        r'solution : version projects global'
        desc = t[3]
        if desc is None:
            desc = nxpy.msvs._impl.solution.ast.Description(("", self.desc_pos))
        t[0] = nxpy.msvs._impl.solution.ast.Solution(self.text, t[1], t[2], desc)
    
    def p_version(self, t):
        r'version : MICROSOFT VISUAL STUDIO SOLUTION FILE COMMA FORMAT VERSION NUMBER'
        t[0] = t[9]

    def p_projects(self, t):
        r'''projects : projects project 
                       | project'''
        if len(t) == 3:
            t[0] = t[1] + [ t[2] ]
        else:
            t[0] = [ t[1] ]
    
    def p_project(self, t):
        r'project : PROJECT LPAREN STRING RPAREN EQUALS project_def ENDPROJECT'
        t[0] = t[6]
    
    def p_project_def(self, t):
        r'project_def : STRING COMMA STRING COMMA STRING project_section'
        t[0] = nxpy.msvs._impl.solution.ast.Project(t[1], t[3])
    
    def p_project_section(self, t):
        r'''project_section : PROJECTSECTION tokens ENDPROJECTSECTION
                              | empty'''
        pass

    def p_global(self, t):
        r'global : GLOBAL global_sections ENDGLOBAL'
        t[0] = t[2]
    
    def p_global_sections(self, t):
        r'''global_sections : global_sections global_section
                              | global_section'''
        if t[1] is not None:
            t[0] = t[1]
        elif len(t) == 3:
            t[0] = t[2]
    
    def p_global_section(self, t):
        r'''global_section : GLOBALSECTION LPAREN IDENTIFIER RPAREN EQUALS IDENTIFIER property_list ENDGLOBALSECTION'''
        t[0] = t[7]
        if t[3][0] == "SolutionConfigurationPlatforms":
            self.desc_pos = t[8][1] - len(t[8][0]) - 1

    def p_property_list(self, t):
        r'''property_list : properties DESCRIPTION
                            | properties'''
        if len(t) == 3:
            t[0] = nxpy.msvs._impl.solution.ast.Description(t[2])
            
    def p_tokens(self, t):
        r'''tokens : anything
                     | tokens anything'''
        if len(t) == 3:
            t[0] = " ".join([ t[1], t[2] ])
        else:
            t[0] = t[1]
    
    def p_anything(self, t):
        r'''anything : COMMA
                       | EQUALS
                       | LBRACE
                       | LPAREN
                       | POINT
                       | RBRACE
                       | RPAREN
                       | VBAR
                       | NUMBER
                       | IDENTIFIER
                       | STRING
                       | UUID
                       | ANY'''
        t[0] = t[1]

    def p_properties(self, t):
        r'''properties : properties property
                         | property'''
        pass

    def p_property(self, t):
        r'''property : term EQUALS term'''
        pass
                
    def p_term(self, t):
        r'''term : build VBAR platform
                   | IDENTIFIER'''
        pass
        
    def p_build(self, t):
        r'''build : LBRACE UUID RBRACE POINT build_type
                    | build_type'''
        pass
    
    def p_build_type(self, t):
        r'''build_type : DEBUG
                         | release'''
        pass
    
    def p_platform(self, t):
        r'''platform : ANY CPU details
                       | ANY CPU'''
        pass
    
    def p_details(self, t):
        r'''details : details POINT detail
                      | POINT detail'''
        pass
    
    def p_detail(self, t):
        r'''detail : DEBUG
                     | release
                     | IDENTIFIER
                     | NUMBER'''
        pass
    
    def p_release(self, t):
        r'''release : RELEASE WITH CODE ANALYSIS
                      | RELEASE'''
        pass

    def p_empty(self, t):
        'empty :'
        pass

    def p_error(self, t):
        _log.error("Syntax error: line (" + str(t.lineno) + ") " + t.type + " = " + t.value)

    def parse(self, text):
        self.text = text
        text = text.split('\n', 1)
        return super(Parser, self).parse(text[1])



# nxpy.maven package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Maven artifacts.

"""

from __future__ import absolute_import

import functools
import os.path

import nxpy.xml.util


_ns = nxpy.xml.util.Namespace("http://maven.apache.org/POM/4.0.0")


def _trim(s):
    return s.replace("-", "").lower()

@functools.total_ordering
class Artifact(object):
    def __init__(self, element):
        self._modified = False
        self._groupId = _ns.find(element, "groupId")
        self._artifactId = _ns.find(element, "artifactId")
        self._version = _ns.find(element, "version")
        try:
            self._packaging = _ns.find(element, "packaging").text
        except AttributeError:
            try:
                self._packaging = _ns.find(element, "type").text
            except AttributeError:
                self._packaging = None
        if self._packaging is None:
            self._packaging = "jar"
        try:
            self._relativePath = os.path.normpath(_ns.find(element, "relativePath").text)
        except AttributeError:
            self._relativePath = None

    groupId = nxpy.xml.util.make_property("_groupId")
    artifactId = nxpy.xml.util.make_property("_artifactId")
    version = nxpy.xml.util.make_property("_version")
    
    def qualified_name(self, full=False):
        name = self.groupId + ":" + self.artifactId
        if full:
            name += ":" + self.packaging + ":" + self.version
        return name

    def before(self, other):
        return ( ( [ _trim(s) for s in self.groupId.split(".") ], _trim(self.artifactId) ) < 
                ( [ _trim(s) for s in other.groupId.split(".") ], _trim(other.artifactId) ) )
        
    @property
    def modified(self):
        return self._modified

    @property
    def packaging(self):
        return self._packaging
    
    @property
    def relativePath(self):
        return self._relativePath
    
    def __str__(self):
        v = ""
        try:
            v = self.version
        except AttributeError:
            pass
        return "%s:%s:%s" % ( self.groupId, self.artifactId, v )

    def __eq__(self, value):
        return ( ( self.groupId, self.artifactId ) == ( value.groupId, value.artifactId ) )

    def __lt__(self, value):
        return ( ( self.groupId, self.artifactId ) < ( value.groupId, value.artifactId ) )

    def __hash__(self):
        return hash("%s:%s" % ( self.groupId, self.artifactId ))



# nxpy.command package -------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Non interactive command driver.

"""

from __future__ import absolute_import

import logging
import os
import subprocess

import nxpy.core._impl.log


_log = logging.getLogger(__name__)


class Error(Exception):
    r"""Raised when command execution fails."""
    def __init__(self, cmd, returncode, err):
        r"""Takes the command line, the error code and the contents of the error stream."""
        self.cmd = cmd
        self.returncode = returncode
        self.stderr = err
        super(Error, self).__init__("%s(%d):\n%s" % ( cmd, returncode, err))


class Command(object):
    r"""
    Represents the command to be executed.
    Typically you would derive from this class and provide a different method for each alternative
    way of invoking the program. If the program you want to execute has many sub-commands you
    might provide a different method for each sub-command. You can use the 
    :py:class:`.option.Config` class to declare the options supported by your command and then use
    the :py:class:`.option.Parser` class to validate your methods' arguments and generate the
    resulting command line. A debug mode is available in which commands are echoed rather than run.
    This can be enabled globally or separately for each invocation.
    
    """
    def __init__(self, cmd, debug=False):
        r"""
        Takes as arguments the command name and a boolean value indicating whether debug mode
        should be activated for all executions of this command.
        
        """
        self.cmd = cmd
        self.debug = debug
    
    def run(self, parser, debug=False):
        r"""
        Executes the command.
        Takes as arguments a command line parser (see the *option* module) and a boolean
        indicating whether debug mode should be used for this execution.
        
        """
        os.environ["LANG"] = "C"
        cmdline = []
        if debug or self.debug:
            cmdline.append("echo")
        cmdline.append(self.cmd)
        cmdline.append(parser.getCommandLine())
        cmd = " ".join(cmdline)
        _log.debug(cmd)
        popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, 
                                 universal_newlines=True)
        out, err = popen.communicate(None)
        _log.debug("----- OUTPUT -----")
        _log.debug(out)
        _log.debug("----- ERROR -----")
        _log.debug(err)
        if popen.returncode != 0:
            raise Error(cmd, popen.returncode, err)
        return out, err



# nxpy.scons package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
SCons Actions to be performed when cleaning.

"""

from __future__ import absolute_import

import nxpy.scons.util


def clean_action(env, targets, action):
    r"""
    To be installed with:
    env.AddMethod(nxpy.scons.clean_action.clean_action, 'CleanAction')
    """
    if nxpy.scons.util.is_cleaning(env, targets):
            env.Execute(action)



# nxpy.maven package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Manipulation of Maven Assembly Plugin Assembly Descriptor files.

"""

from __future__ import absolute_import

import os.path

import nxpy.xml.util


class BadAssemblyDescriptorFileError(EnvironmentError):
    pass


class AssemblyDescriptor(object):
    
    def __init__(self, path):
        self.path = os.path.realpath(path)
        self.dir = os.path.split(self.path)[0]
        try:
            self.tree = nxpy.xml.util.parse(path)
        except IOError:
            raise BadAssemblyDescriptorFileError(path)
        self.root = self.tree.getroot()
        self._ns = nxpy.xml.util.Namespace(element=self.root)

        depset = self._ns.find(self.root, "dependencySets")
        deps = None
        if depset is not None:
            if len(depset) > 0:
                deps = depset[0]
            else:
                deps = self._ns.SubElement(depset, "dependencySet")
        if deps is None:
            raise BadAssemblyDescriptorFileError(path)
        self.includes = nxpy.xml.util.SequenceElement(deps, "includes", "include", self._ns.url)
        self.excludes = nxpy.xml.util.SequenceElement(deps, "excludes", "exclude", self._ns.url)
        self._writer = nxpy.xml.util.Writer(self._make_root_tag(), {}, 4)

    _schema_loc = nxpy.xml.util.QName(
            r"{http://www.w3.org/2001/XMLSchema-instance}schemaLocation")
    
    _root_fmt = ( 
r'''<{0}
    xmlns="{1}"
    xmlns:xsi="{2}"
    xsi:{3}="{4}">'''
        )
    
    def _make_root_tag(self):
        root = nxpy.xml.util.QName(self.root.tag)
        return AssemblyDescriptor._root_fmt.format(root.tag, root.url, 
                AssemblyDescriptor._schema_loc.url, AssemblyDescriptor._schema_loc.tag,
                self.root.get(AssemblyDescriptor._schema_loc.text))
    
    @property
    def modified(self):
        return ( ( self.includes and self.includes.modified ) or 
                 ( self.excludes and self.excludes.modified ) )

    def write(self, where):
        if not where:
            where = self.path
        self._writer.write(self.root, where)

    def save(self):
        if self.modified:
            self.write(None)
            self.includes.modified = False
            self.excludes.modified = False
            return True
        return False



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2015
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Backup a file or directory to make editing reversible.

Implement the context manager protocol, so as to be suitable to be used with
the *with* statement. When used in this fashion changes are discarded when an 
exception is thrown.

"""

from __future__ import absolute_import

import os.path
import shutil
import tempfile

import six

import nxpy.core.error
import nxpy.core.file_object
import nxpy.core.path


class NotSavedError(Exception):
    r"""
    Raised when commit or rollback is called on an inactive *BackUpFile* or *BackUpDirectory*.
    
    """


class SaveError(Exception):
    r"""Raised when a backup file or directory could not be created."""


class RemovalError(Exception):
    r"""Raised to signal errors in the removal of backup files or directories."""


class MissingBackupError(Exception):
    r"""raised when a backup file or directory isn't found."""


class BackupFile(nxpy.core.file_object.ReadOnlyFileObject):
    r"""
    Implements a read only file object used to automatically back up a file that has to be
    modified.
    
    """
    _prefix = "_BackupFile"

    MOVE = 1
    COPY = 2
    
    def __init__(self, file_, ext=".BAK", dir=".", mode=COPY):
        r"""
        Prepare to backup *file_*, either a file-like object or a path. 
        
        The backup file will be created in directory *dir* with extension *ext*. If *mode* is 
        *COPY* the original file will be copied to the backup destination; if *mode* is *MOVE* it 
        will be moved there.
        
        """
        super(BackupFile, self).__init__()
        if mode not in ( BackupFile.MOVE, BackupFile.COPY ):
            raise nxpy.core.error.ArgumentError(self.mode + ": Invalid mode")
        if isinstance(file_, six.string_types):
            self._orig_name = file_
            self._orig_file = None
            self._bck_name = os.path.join(dir, self._orig_name) + ext            
        else:
            self._orig_name = None
            self._orig_file = file_
            self._bck_name = tempfile.mktemp(ext, BackupFile._prefix, dir)
        if mode == BackupFile.MOVE and self._orig_name is None:
            raise nxpy.core.error.ArgumentError("Mode can only be MOVE when file_ is a path.")
        self._bck_file = None
        self._mode = mode
        self._saved = False

    @property
    def name(self):
        r"""The name of the file to be backed up."""
        if self._orig_name is not None:
            return self._orig_name
        else:
            return self._orig_file.name

    def __enter__(self):
        r"""When the controlling *with* statement is entered, create the backup file."""
        self.save()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        r"""
        When the controlling *with* statement is exited normally discard the backup file,
        otherwise restore it to its original place.
        
        """
        if self._saved:
            if exc_type is None:
                self.commit()
            else:
                self.rollback()
        return False

    def save(self):
        r"""
        Create a backup copy of the original file.
        
        Throw *SaveError* if it wasn't possible.
        """
        self.close()
        if self._mode == BackupFile.MOVE:
            try:
                shutil.move(self._orig_name, self._bck_name)
            except:
                raise SaveError()
        elif self._mode == BackupFile.COPY:
            try:
                if self._orig_file is None:
                    self._orig_file = open(self._orig_name, "r+")
                    self._tell = self._orig_file.tell()
                else:
                    self._tell = self._orig_file.tell()
                    self._orig_file.seek(0, os.SEEK_SET)
                self._bck_file = open(self._bck_name,"w+")
                shutil.copyfileobj(self._orig_file, self._bck_file, -1)
                self._orig_file.seek(self._tell, os.SEEK_SET)
                self._bck_file.close()
                self._bck_file = None
            except:
                raise SaveError("Could not backup file " + self._orig_name
                        if self._orig_name is not None else self._bck_name)
        self._saved = True

    def commit(self):
        r"""Discard the backup, i.e. keep the supposedly modified file."""
        if not self._saved:
            raise NotSavedError(self._orig_name + ": File not saved")
        self.close()
        try:
            os.remove(self._bck_name)
        except:
            raise RemovalError("Error removing file " + self._bck_name)
        self._saved = False

    def rollback(self):
        r"""Replace the original file with the backup copy."""
        if not self._saved:
            raise NotSavedError(self._orig_name + ": File not saved")
        self.close()
        try:
            if self._mode == BackupFile.MOVE:
                shutil.move(self._bck_name, self._orig_name)
            elif self._mode == BackupFile.COPY:
                self._bck_file = open(self._bck_name,"r")
                self._orig_file.seek(0, os.SEEK_SET)
                shutil.copyfileobj(self._bck_file, self._orig_file, -1)
                if self._orig_name is None:
                    self._orig_file.seek(self._tell, os.SEEK_SET)
                else:
                    self._orig_file.close()
                self._bck_file.close()
                self._bck_file = None
                os.remove(self._bck_name)
        except:
            raise MissingBackupError("File " + self._bck_name + 
                        " not found")
        self._saved = False

    BINARY = 3
    TEXT = 4
    
    def open(self, mode=TEXT):
        r"""Open the backup file for reading. *mode* may be either *TEXT* or *BINARY*."""
        if not self._saved:
            raise NotSavedError(self._name + ": File not saved")
        if mode == BackupFile.TEXT:
            self._bck_file = open(self._bck_name,"r")
        elif mode == BackupFile.BINARY:
            self._bck_file = open(self._bck_name,"rb")
        self.setFile(self._bck_file)

    def close(self):
        r"""
        Close the backup file and release the corresponding reference.
        
        The backup file may not be reopened.
        
        """
        if self._bck_file:
            self._bck_file.close()
            self._bck_file = None
            self.setFile(None)


class BackupDir(object):
    r"""
    Move or copy a directory that needs to be recreated or modified.
    
    """
    MOVE = 1
    COPY = 2
    
    def __init__(self, dir_, ext=".BAK", mode=MOVE):
        r"""
        Prepare to backup the *dir_* directory. 
        
        The backup will be created in *dir_*'s parent directory, which must be writable, with 
        extension *ext*. If *mode* is *MOVE*, the default, the original directory will be moved to
        the backup destination; if *mode* is *COPY* it will be copied there.
        
        """
        if mode not in ( BackupDir.MOVE, BackupDir.COPY ):
            raise nxpy.core.error.ArgumentError(self.mode + ": Invalid mode")
        self._orig_dir = dir_
        self._bck_dir = dir_ + ext
        self._mode = mode
        self._saved = False

    def __enter__(self):
        r"""When the controlling *with* statement is entered, create the backup directory."""
        self.save()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        r"""
        When the controlling *with* statement is exited normally discard the backup directory,
        otherwise restore it to its original place.
        
        """
        if self._saved:
            if exc_type is None:
                self.commit()
            else:
                self.rollback()
        return False

    def save(self):
        r"""Create a backup copy of the original directory."""
        if self._mode == BackupFile.MOVE:
            try:
                shutil.move(self._orig_dir, self._bck_dir)
            except:
                raise SaveError("Could not backup directory " + self._orig_dir)
        elif self._mode == BackupFile.COPY:
            try:
                shutil.copytree(self._orig_dir, self._bck_dir)
            except:
                raise SaveError("Could not backup directory " + self._orig_dir)
        self._saved = True

    def commit(self):
        r"""Discard the backup, i.e. keep the supposedly modified file."""
        if not self._saved:
            raise NotSavedError(self._orig_dir + ": directory not saved")
        try:
            nxpy.core.path.blasttree(self._bck_dir)
        except:
            raise RemovalError("Error removing directory " + self._bck_dir)
        self._saved = False

    def rollback(self):
        r"""Replace the original file with the backup copy."""
        if not self._saved:
            raise NotSavedError(self._orig_dir + ": directory not saved")
        try:
            temp_dir = None
            if os.path.isdir(self._orig_dir):
                temp_dir = self._orig_dir + ".NEW"
                shutil.copytree(self._orig_dir, temp_dir)
            nxpy.core.path.blasttree(self._orig_dir)
            shutil.copytree(self._bck_dir, self._orig_dir)
            if temp_dir is not None:
                nxpy.core.path.blasttree(temp_dir)
        except:
            raise MissingBackupError("Directory " + self._bck_dir + " not found")
        self._saved = False



# nxpy.scons package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Utilities for the SCons build tool.

"""

from __future__ import absolute_import

import os

from six.moves import range


# Taken almost verbatim from this SCons Wiki recipe: http://www.scons.org/wiki/CustomCleanActions
# (If you are the original author and object to this being here, get in touch!)
 
def is_cleaning(env, targets):
        if not env.GetOption('clean'):
                return False
        # normalize targets to absolute paths
        targets = env.subst('${TARGETS.abspath}', target=targets).split()
        launchdir = env.GetLaunchDir()
        topdir = env.Dir("#").abspath
        cl_targets = env.subst('$COMMAND_LINE_TARGETS').split()
        if len(cl_targets) == 0:
                cl_targets.append(".")
        for i in range(len(cl_targets)):
                full_target = ""
                if cl_targets[i][0] == '#':
                        full_target = os.path.join(topdir,cl_targets[i][1:0])
                else:
                        full_target = os.path.join(launchdir,cl_targets[i])
                full_target = os.path.normpath(full_target)
                for target in targets:
                        if target.startswith(full_target):
                                return True
        return False



# nxpy.maven package ---------------------------------------------------------

# Copyright Nicola Musatti 2011 - 2012
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Tools to drive the Maven build tool and to manipulate its configuration files.

"""



# nxpy.ply package -----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Class wrapper for PLY's yacc module.

"""

from __future__ import absolute_import

import ply.yacc


class Parser(object):
    def __init__(self, scanner, debug=False):
        self.scanner = scanner
        self.debug = debug
        Parser.tokens = self.scanner.tokens
        self.parser = ply.yacc.yacc(module=self, debug=self.debug)

    def parse(self, input_):
        self.scanner.reset(input_)
        return self.parser.parse(lexer=self.scanner.lexer)



# nxpy package -------------------------------------------------------

# Copyright Nicola Musatti 2011 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Nxpy is an etherogeneous collection of useful modules, dealing with diverse topics such as 
wrapping complex commands with API's or interacting with Java and .NET build systems. 
There are also more basic utilities, such as automation of backup files, support for writing your 
own file-like objects and many other things.

The library is being developed with Python 3.6 so as to be compatible with Python 2.7. Tests are run
and most modules work also with 2.6, 3.3, 3.4 and 3.5. Some should still work with 3.2 and 2.5.

.. IMPORTANT::
   Nxpy is a personal project and, while I use a good part of its contents on an almost daily basis,
   my use cases are too specific to guarantee that everything works as expected.

The Nxpy logo was drawn by Claudia Romano.

"""



# nxpy.wx package ------------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

from __future__ import absolute_import

import wx

import nxpy.wx.interpreter


_app = None

class InterpreterTester(nxpy.wx.interpreter.Interpreter):
    r"""
    Test oriented Interpreter implementation.
    
    Useful for testing wxPython based command interpreters isolated from the visual infrastructure.
    """
    def __init__(self, cmd):
        global _app
        if not _app:
            _app = wx.App()
        super(InterpreterTester, self).__init__(cmd)


# nxpy.ccase package ---------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

"""
Programming interface to the ClearCase configuration management tool.
Implemented by driving interactively the cleartool program.

"""

from __future__ import absolute_import

import re
import sys

import nxpy.command.interpreter
import nxpy.command.option


class ClearToolError(Exception):
    r"""Raised when cleartool returns an unexpected reply."""


class InvalidArgument(ClearToolError):
    r"""Raised on an invalid argument combination."""


class FailedCommand(nxpy.command.interpreter.BadCommand, ClearToolError):
    r"""Raised when a command returns an error code."""

    def __init__(self, cmd, err_code=0, err=""):
        message = []
        if err:
            message.append(err)
        if err_code != 0:
            message.append("Error: " + err_code)
        super(FailedCommand, self).__init__(cmd, "\n".join(message))


_config = nxpy.command.option.Config(

    prefix="-",

    # boolean options, which appear on the command line without arguments.
    bool_opts=("cact", "cview", "force", "identical", "keep", "long", "me",
            "none", "nxn", "preview", "print_report", "short", "slink",
            "visible", "vob_only"),

    # Options with an associated value
    value_opts=("activity", "comment", "in_stream", "invob", "log", "proj",
            "stream", "target", "to", "user", "view"),

    # Options with multiple arguments, separated by commas.
    iterable_opts=("activities", ),

    # Options whose argument is copied by means of the '%' operator.
    format_opts={"fmt": "\"%s\""},

    # Options whose name must be translated, e.g. because it is not a valid identifier.
    mapped_opts={"in_stream": "-in", "nolog": "-log NUL", "print_report": "-print", "proj": "-in"},

    # Boolean options that are expressed by the lack of an opposite command line option
    opposite_opts={"activity": "-none", "checkout": "-nco -force", "comment": "-nc", "keep": "-rm",
            "log": "-log NUL"}

    )


command_line = "cleartool -status"


class ClearTool(object):
    r"""
    Allows manipulation of ClearCase UCM projects by driving the  cleartool utility in a
    subprocess.
    
    Each sub-command is implemented in its own method. In general cleartool's standard output is 
    returned, except when standard error contains useful information, in which case both are 
    returned.
    
    """
    _result_re = re.compile(r"Command \d+ returned status (\d)\r\n")

    def __init__(self, cmd=None, log=False):
        r"""
        Create a *cleartool* interpreter.
        
        *cmd* is an optional :py:class:`.command.interpreter.Interpreter` instance, used mainly to
        supply an alternative implementation for tests; *log* is an optional logging destination.
        
        """
        if cmd is not None:
            self.cmd = cmd
        else:
            self.cmd = nxpy.command.interpreter.Interpreter(command_line)
        self.cmd.setLog(log)

    def _run(self, parser, **kwargs):
        if isinstance(parser, nxpy.command.option.Parser):
            cmd = parser.getCommandLine()
        else:
            cmd = parser
        raise_on_failure = False
        try:
            raise_on_failure = kwargs["raise_on_failure"]
            del kwargs["raise_on_failure"]
        except KeyError:
            pass
        try:
            kwargs["cond"] = nxpy.command.interpreter.RegexpWaiter(self._result_re,
                    nxpy.command.interpreter.EXP_OUT)
            out, err = self.cmd.run(cmd, **kwargs)
            if raise_on_failure:
                err_code = ClearTool._result_re.search(out).group(1)
                if err_code > 0:
                    raise FailedCommand(cmd, err_code=err_code)
            return ClearTool._result_re.sub("", out), err, cmd
        except nxpy.command.interpreter.BadCommand:
            e = sys.exc_info()[1]
            raise FailedCommand(e.command, err=e.stderr)

# The cleartool sub-commands

    def cd(self, dir_=""):
        self._run("cd " + dir_)

    def checkin(self, *elements, **options):
        if not elements:
            raise InvalidArgument("At least one element must be specified")
        op = nxpy.command.option.Parser(_config, "checkin", elements, options, comment="", 
                identical=False)
        return self._run(op)[0]

    def checkout(self, *elements, **options):
        if not elements:
            raise InvalidArgument(
                    "At least one element must be specified")
        op = nxpy.command.option.Parser(_config, "checkout -nq", elements, options, comment="")
        return self._run(op)[0]

    def deliver(self, **options):
        op = nxpy.command.option.Parser(_config, "deliver", (), options, activities=(), cact=False,
                long=False, preview=False, short=False, stream="", target="",
                to="")
        op.checkExclusiveOptions("activities", "cact")
        op.checkExclusiveOptions("long", "short")
        return self._run(op, raise_on_error=False)

    def describe(self, *args, **options):
        op = nxpy.command.option.Parser(_config, "describe", args, options, fmt="", short=False)
        op.checkExclusiveOptions("fmt", "short")
        return self._run(op)[0]

    def ln(self, dest, *src, **options):
        r"""Note: Arguments are inverted with respect to the original command, beware!"""
        args = src + (dest, )
        op = nxpy.command.option.Parser(_config, "ln", args, options, checkout=True, comment=False, 
                slink=True)
        return self._run(op)[0]

    def ls(self, *files, **options):
        op = nxpy.command.option.Parser(_config, "ls", files, options, nxn=True, short=True, 
                visible=False, vob_only=False)
        return self._run(op)[0]
    
    def lsactivity(self, *activities, **options):
        r"""
        Note: *fmt=r"%[versions]p\n"* causes a race condition with activities that have a large
        number of contribuents.

        """
        op = nxpy.command.option.Parser(_config, "lsactivity", activities, options, cact=False, 
                fmt="", long=False, me=False, short=False, in_stream="", 
                user="", view="")
        op.checkExclusiveOptions("in_stream", "cact", "user", "me")
        op.checkExclusiveOptions("fmt", "long", "short")
        return self._run(op, interval=0.1)[0]
    
    def lshistory(self, obj, **options):
        op = nxpy.command.option.Parser(_config, "lshistory", ( obj, ), options, fmt="")
        return self._run(op)[0]

    def lsproject(self, **options):
        op = nxpy.command.option.Parser(_config, "lsproject", (), options, cview=False, view="", 
                invob="", fmt="")
        op.checkExactlyOneOption("cview", "view", "invob")
        return self._run(op)[0]

    def lsstream(self, **options):
        op = nxpy.command.option.Parser(_config, "lsstream", (), options, view="", proj="", 
                invob="", fmt="")
        op.checkExclusiveOptions("view", "proj", "invob")
        return self._run(op)[0]

    def lsview(self, *tags, **options):
        op = nxpy.command.option.Parser(_config, "lsview", [ "\'%s\'" % (t) for t in tags ],
                options, cview=False, short=False)
        op.checkNotBothOptsAndArgs("cview")
        return self._run(op)[0]

    def lsvob(self):
        op = nxpy.command.option.Parser(_config, "lsvob", (), {})
        return self._run(op)[0]

    def mv(self, dest, *src, **options):
        r"""Note: Arguments are inverted with respect to the original command, beware!"""
        args = src + (dest, )
        op = nxpy.command.option.Parser(_config, "mv", args, options, comment="")
        return self._run(op)[0]

    def pwd(self):
        return self._run("pwd")[0].strip()

    def rmname(self, *args, **options):
        op = nxpy.command.option.Parser(_config, "rmname", args, options, checkout=True, comment="")
        return self._run(op)[0]
        
    def setactivity(self, activity, **options):
        if activity:
            args = ( activity, )
        else:
            args = ()
            options["none"] = True
        op = nxpy.command.option.Parser(_config, "setactivity", args, options, none=False, view="")
        op.checkOneBetweenOptsAndArgs("none")
        return self._run(op)[0]

    def uncheckout(self, *elements, **options):
        op = nxpy.command.option.Parser(_config, "uncheckout", elements, options, keep=False)
        return self._run(op)[0]

    def update(self, **options):
        op = nxpy.command.option.Parser(_config, "update", (), options, print_report=False,
                force=False, nolog=False, log="")
        op.checkExclusiveOptions("log", "nolog")
        return self._run(op, raise_on_error=False, timeout=300, interval=0.5, quantum=0.5)[0]



# nxpy.svn package -----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Subversion URL manipulation.

"""

from __future__ import absolute_import

from six.moves import urllib


class BadUrlError(EnvironmentError):
    r"""Indicates a malformed URL."""


class Url(object):
    r"""
    A well-formed Subversion repository URL that follows standard svn conventions.
    
    The URL must end in either 'trunk', 'tags/label' or 'branches/label'.
    
    """
    def __init__(self, path):
        url = urllib.parse.urlsplit(path)
        if url is None:
            raise BadUrlError(path)
        self.url = url
        path = url.path.split("/")
        if path[-1] == "trunk":
            self.trunk = True
            self.name = path[-2]
        else:
            self.trunk = False
            self.name = path[-3]
        self.tag = None
        self.branch = None
        if path[-2] == "tags":
            self.tag = path[-1]
            self.branch = None
        if path[-2] == "branches":
            self.branch = path[-1]
        if not self.trunk and not self.tag and not self.branch:
            raise BadUrlError(path)

    def __eq__(self, other):
        return self.url == other.url
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __str__(self):
        return self.url.geturl()

    def _urlreplace(self, old, new):
        return Url(self.url.geturl().replace(old, new))

    def istrunk(self):
        return self.trunk
    
    def istag(self, tag=None):
        return ( not tag and self.tag ) or ( tag and self.tag == tag )
    
    def isbranch(self, branch=None):
        return ( not branch and self.branch ) or ( branch and self.branch == branch )
    
    def gettrunk(self):
        if self.trunk:
            return self
        new = "trunk"
        if self.tag:
            old = "tags/" + self.tag
        elif self.branch:
            old = "branches/" + self.branch
        return self._urlreplace(old, new)

    def gettag(self, tag):
        if self.tag:
            if self.tag == tag:
                return self
            else:
                old = "tags/" + self.tag
        new = "tags/" + tag
        if self.trunk:
            old = "trunk"
        elif self.branch:
            old = "branches/" + self.branch
        return self._urlreplace(old, new)

    def getbranch(self, branch):
        if self.branch:
            if self.branch == branch:
                return self
            else:
                old = "branches/" + self.branch
        new = "branches/" + branch
        if self.trunk:
            old = "trunk"
        elif self.tag:
            old = "tags/" + self.tag
        return self._urlreplace(old, new)



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2013 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Helpers for the standard :py:mod:`abc` module.

"""

from __future__ import absolute_import

import nxpy.core.past


if nxpy.core.past.V_2_7.at_most():
    
    class abstractstatic(staticmethod):
        r"""
        Decorator that combines :py:obj:`staticmethod` and :py:obj:`abc.abstractmethod`.
        
        Copied from `this answer <http://stackoverflow.com/a/4474495/838975>`_ to `this StackOverflow
        question
        <http://stackoverflow.com/questions/4474395/staticmethod-and-abc-abstractmethod-will-it-blend>`_.
        
        """
        __slots__ = ()
        
        def __init__(self, function):
            super(abstractstatic, self).__init__(function)
            function.__isabstractmethod__ = True
    
        __isabstractmethod__ = True

elif nxpy.core.past.V_3_2.at_least():
    
    import abc
    
    abstractstatic = abc.abstractstaticmethod
    


# nxpy.core._impl package ----------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
This module should be imported in all modules that use the standard logging module

"""

from __future__ import absolute_import

import logging

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

logging.getLogger("nxpy").addHandler(NullHandler())



# nxpy.test package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Unittest utility functions.

"""

from __future__ import absolute_import

import sys

import nxpy.core.past


if nxpy.core.past.V_2_6.at_most():
    import unittest2
    TestCase = unittest2.TestCase
    TestLoader = unittest2.TestLoader
    TestSuite = unittest2.TestSuite
    TextTestRunner = unittest2.TextTestRunner
    main = unittest2.main
    skip = unittest2.skip
else:
    import unittest
    TestCase = unittest.TestCase
    TestLoader = unittest.TestLoader
    TestSuite = unittest.TestSuite
    TextTestRunner = unittest.TextTestRunner
    main = unittest.main
    skip = unittest.skip


def testModules(*modules):
    r"""Runs all tests defined in the given *modules*."""
    loader = TestLoader()
    loader.sortTestMethodsUsing = None
    suite = TestSuite()
    for m in modules:
        suite.addTests(loader.loadTestsFromModule(sys.modules[m]))
    TextTestRunner(verbosity=2).run(suite)


def testClasses(*classes):
    r"""Runs all tests defined in the given *classes*."""
    loader = TestLoader()
    loader.sortTestMethodsUsing = None
    suite = TestSuite()
    for c in classes:
        suite.addTests(loader.loadTestsFromTestCase(c))
    TextTestRunner(verbosity=2).run(suite)


def skipIfNotAtLeast(version):
    r"""Skip the current test if the current Python release is lower than *version*."""
    if version.at_least():
        return lambda func: func
    return skip("Requires at least Python " + str(version))


def skipIfNotAtMost(version):
    r"""Skip the current test if the current Python release is higher than *version*."""
    if version.at_most():
        return lambda func: func
    return skip("Requires at most Python " + str(version))



# nxpy.core package ----------------------------------------------------

# Copyright Nicola Musatti 2013 - 2015
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
filesystem related utilities.

"""

from __future__ import absolute_import

import logging
import os
import shutil
import stat


def _on_error(func, path, ex_info):
    r"""Error handler for :py:func:`shutil.rmtree`."""
    try:
        mode = os.stat(path).st_mode
        if not stat.S_ISLNK(mode):
            os.chmod(path, mode | stat.S_IWRITE)
            if stat.S_ISDIR(mode):
                os.rmdir(path)
            else:
                os.remove(path)
        return
    except:
        pass
    logging.error(path + ": not removed")


def blasttree(dir_):
    r"""Remove a directory more stubbornly than :py:func:`shutil.rmtree`.

    Required on filesystems that do not allow removal of non-writable files
    
    """
    shutil.rmtree(dir_, ignore_errors=False, onerror=_on_error)


class CurrentDirectory(object):
    r"""A context manager that allows changing the current directory temporarily."""
    def __init__(self, path):
        r"""Set the current directory to *path*."""
        self._path = path
        self._prev = os.getcwd()
        self._curr = self._prev
    
    @property
    def current(self):
        r"""Return the current directory."""
        return self._curr

    def __enter__(self):
        os.chdir(self._path)
        self._curr = os.getcwd()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._prev)
        self._curr = self._prev
        return False



# nxpy.command package -------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Function argument to command line option conversion.
Provides means to describe commands with complicated syntaxes, which often combine sub-commands, 
options and arguments. Typical examples include subversion and ftp.

"""

from __future__ import absolute_import

import nxpy.core.sequence


class InvalidOptionError(Exception):
    r"""Raised when an option is not supported."""


class Config(object):
    r"""
    Command option definitions.
    Provides a single definition point for all the options supported by a command. 
    
    """
    def __init__(self, prefix="--", separator=" ", bool_opts=(), value_opts=(), iterable_opts=(), 
                 format_opts={}, mapped_opts={}, opposite_opts={}):
        r"""
        Constructor. Its arguments are used to specify all the valid options.
        Each option is prefixed by *prefix*. When an option takes multiple arguments these are
        separated by a *separator*. *bool_opts* must be specified on the command line when they
        are *True*. *value_opts* take a single argument; *iterable_opts* take multiple arguments;
        *format_opts* have their syntax specified by means of a format string;
        *mapped_opts* require some form of translation, usually because they are not valid
        Python identifiers; *opposite_opts* must be specified on the command line when they
        are *False*.
        
        """
        self.prefix = prefix
        self.separator = separator
        self.bool_opts = bool_opts
        self.value_opts = value_opts
        self.iterable_opts = iterable_opts
        self.format_opts = format_opts
        self.mapped_opts = mapped_opts
        self.opposite_opts = opposite_opts
        
        self.opts = set(nxpy.core.sequence.make_tuple(bool_opts))
        self.opts.update(nxpy.core.sequence.make_tuple(value_opts))
        self.opts.update(nxpy.core.sequence.make_tuple(iterable_opts))
        self.opts.update(format_opts.keys())
        self.opts.difference_update(mapped_opts.keys())


class Parser(object):
    r"""
    Constructs a complex command line from the provided *command* and its *options* and
    *arguments*. Uses a :py:class:`.Config` instance, *config*, to provide means to check conditions
    on the supplied options. Other constraints on how options should be used may be expressed and
    verified by means of the *check* methods.
    
    """
    def __init__(self, config, command, arguments, options, **defaults):
        r"""
        Takes an instance of :py:class:`.Config`, a *command* to execute, an iterable of *arguments*
        and a mapping of *options* and their actual values. The remaining keyword arguments indicate
        the options supported by *command* with their default values. 
        
        """
        invalid = set(options.keys()).difference(defaults.keys())
        if len(invalid) > 0:
            raise InvalidOptionError(", ".join(invalid) + 
                    ": invalid option(s)")
        self.config = config
        self.command = command
        self.arguments = arguments
        self.options = {}
        self.options.update(defaults)
        self.options.update(options)
        self.cmd_line = []

    def getCommandLine(self):
        r"""Returns the command line to be executed."""
        if not self.cmd_line:
            self._applyOptions()
            if self.arguments:
                self.cmd_line.extend(self.arguments)
        return " ".join(self.cmd_line)

    def checkMandatoryOptions(self, *options):
        r"""Checks that all compulsory options have been specified."""
        if not all([ bool(self.options[o]) for o in options ]):
            raise InvalidOptionError(", ".join(options) +
                    ": mandatory options")

    def checkExclusiveOptions(self, *options):
        r"""Checks that at most one in a set of mutually exclusive options has been specified."""
        if sum([ bool(self.options[o]) for o in options ]) > 1:
            raise InvalidOptionError(", ".join(options) + 
                    ": mutually exclusive options")

    def checkExactlyOneOption(self, *options):
        r"""
        Checks that one and only one in a set of mutually exclusive options has been specified.
        
        """
        s = sum([ bool(self.options[o]) for o in options ])
        if s != 1:
            raise InvalidOptionError(", ".join(options) + 
                    ": only one among these options should be specified")

    def checkNotBothOptsAndArgs(self, *options):
        r"""
        Checks that options incompatible with arguments haven't been specified if any argument is
        present.
        
        """
        if any([ bool(self.options[o]) for o in options ]) and self.arguments:
            raise InvalidOptionError(", ".join(options) + 
                    ": This/these option(s) is/are invalid when arguments are specified")

    def checkOneBetweenOptsAndArgs(self, *options):
        r"""
        Checks that either at least one in a set of options or some arguments have been specified,
        but not both.
        
        """
        if ( any([ bool(self.options[o]) for o in options ]) + 
                bool(self.arguments) ) != 1:
            raise InvalidOptionError(", ".join(options) + 
                    ": This/these option(s) is/are mutually exclusive with arguments")

    def _applyOptions(self):
        r"""Translates options specified as function parameters into command line arguments."""
        if self.command is not None:
            self.cmd_line.append(self.command)
        for opt in self.options.keys():
            if self.options[opt]:
                if opt in self.config.opts:
                    self.cmd_line.append(self.config.prefix + opt)
                elif opt in self.config.mapped_opts:
                    self.cmd_line.append(self.config.mapped_opts[opt])
                
                if opt in self.config.value_opts:
                    self.cmd_line.append(self.options[opt])
                elif opt in self.config.iterable_opts:
                    self.cmd_line.append(self.config.separator.join(self.options[opt]))
                elif opt in self.config.format_opts.keys():
                    self.cmd_line.append(
                            self.config.format_opts[opt] % self.options[opt])
            elif opt in self.config.opposite_opts.keys():
                self.cmd_line.append(self.config.opposite_opts[opt])



# nxpy.msvs._impl.solution package -------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

from __future__ import absolute_import

import logging

import nxpy.ply.scanner

import nxpy.core._impl.log


_log = logging.getLogger(__name__)


class Scanner(nxpy.ply.scanner.Scanner):
    def __init__(self, *args, **kwargs):
        super(Scanner, self).__init__(args, kwargs)

    reserved = ( 
        'Analysis',
        'Any',
        'Code',
        'CPU',
        'Debug',
        'Description',
        'EndGlobal',
        'EndGlobalSection',
        'EndProject',
        'EndProjectSection',
        'File', 
        'Format',
        'Global',
        'GlobalSection',
        'Microsoft',
        'Project',
        'ProjectSection',
        'Release',
        'Studio', 
        'Solution', 
        'Version', 
        'Visual', 
        'with',
        )
    
    tokens = [ r.upper() for r in reserved ] + [
         'COMMA',
         'EQUALS',
         'LBRACE',
         'LPAREN',
         'POINT',
         'RBRACE',
         'RPAREN',
         'VBAR',
         
         'NUMBER',
         'IDENTIFIER',
         'STRING',
         'UUID',
         ]

    # Punctuation

    t_COMMA     = r','
    t_EQUALS    = r'='
    t_LBRACE    = r'\{'
    t_LPAREN    = r'\('
    t_POINT     = r'\.'
    t_RBRACE    = r'\}'
    t_RPAREN    = r'\)'
    t_VBAR      = r'\|'
    
    t_NUMBER    = r'\d+(?:\.\d+)?'
    
    def t_STRING(self, t):
        r'\"[^\"]*\"'
        t.value = t.value[1:-1]
        return t

    def t_UUID(self, t):
        r'\w{8}(?:-\w{4}){3}-\w{12}'
        return t
    
    _reserved_map = dict([ ( s, s.upper() ) for s in reserved ])

    def t_DESCRIPTION(self, t):
        r'Description[ \t]*=[^\n]*'
        t.value = ( t.value, t.lexer.lexpos )
        return t
    
    def t_IDENTIFIER(self, t):
        r'[A-Za-z_][\w_]*'
        t.type = self._reserved_map.get(t.value,"IDENTIFIER")
        t.value = ( t.value, t.lexer.lexpos )
        return t

    # Whitespace
    
    t_ignore = ' \t'
    
    # Comments

    def t_comment(self, t):
        r'\#.*[\n\r]+'
        t.lineno += t.value.count('\n')

    # Line counter

    def t_newline(self, t):
        r'[\n\r]+'
        t.lineno += t.value.count("\n")

    # Errors

    def t_error(self, t):
        _log.error("Illegal character: '%s'", t.value[0])
        t.lexer.skip(1)



# nxpy.maven package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Manipulation of Maven POM files.

"""

from __future__ import absolute_import

import collections
import os.path

import lxml.etree

import nxpy.xml.util
import nxpy.maven.artifact


_ns = nxpy.xml.util.Namespace("http://maven.apache.org/POM/4.0.0")


class BadPomFileError(EnvironmentError):
    pass


class ArtifactAlreadyPresentError(ValueError):
    pass


class MissingArtifactError(ValueError):
    pass


class ModuleAlreadyPresentError(ValueError):
    pass


class MissingModuleError(ValueError):
    pass


class DependencyIterator(collections.Iterator):
    def __init__(self, deps):
        self.iter = iter(deps.dependencies)

    def __next__(self):
        return next(self.iter)[1]


class Dependencies(object):
    def __init__(self, element):
        self.element = element
        self._modified = False
        self.dependencies = []
        if self.element is not None:
            i = 0
            for d in element.getchildren():
                if d.tag is not lxml.etree.Comment:
                    a = nxpy.maven.artifact.Artifact(d)
                    self.dependencies.append([i, a])
                i += 1

    def _saved(self):
        self._modified = False
        for a in self.dependencies:
            a[1]._modified = False

    @property
    def modified(self):
        return self._modified or any([ a[1]._modified for a in self.dependencies ])
    
    def contains(self, artifact):
        if isinstance(artifact, nxpy.maven.artifact.Artifact):
            group_id = artifact.groupId
            artifact_id = artifact.artifactId
        else:
            group_id, artifact_id = artifact.split(":")[0:2]
        for a in self.dependencies:
            if a[1].groupId == group_id and a[1].artifactId == artifact_id:
                return a[1]
        return None
    
    def add(self, artifact):
        if self.contains(artifact):
            raise ArtifactAlreadyPresentError()
        d = _ns.Element("dependency")
        d.text = self.element.text + "    "
        d.tail = self.element.tail + "    "
        gi = _ns.SubElement(d,"groupId")
        gi.text = artifact.groupId
        gi.tail = d.text
        ai = _ns.SubElement(d,"artifactId")
        ai.text = artifact.artifactId
        ai.tail = d.text
        v = _ns.SubElement(d,"version")
        v.text = artifact.version
        if artifact.packaging != "jar":
            v.tail = d.text
            p = _ns.SubElement(d,"type")
            p.text = artifact.packaging
            p.tail = d.tail + "    "
        else:
            v.tail = d.tail + "    "            
        art = nxpy.maven.artifact.Artifact(d)
        art._modified = True
        index = 0
        for i in range(0, len(self.dependencies) -1):
            if index == 0:
                if self.dependencies[i][1] > art:
                    index = i
            else:
                self.dependencies[i][0] += 1
        if index < len(self.dependencies):
            d.tail += "    "
        else:
            self.element[self.dependencies[index-1][0]].tail += "    "
        self.element.insert(self.dependencies[index][0], d)
        self.dependencies.insert(index, [self.dependencies[index][0], art])

    def remove(self, artifact):
        if not self.contains(artifact):
            raise MissingArtifactError()
        index = -1
        for i in range(0, len(self.dependencies) -1):
            if index == -1:
                if ( self.dependencies[i][1].artifactId == artifact.artifactId and 
                        self.dependencies[i][1].groupId == artifact.groupId ):
                    index = i
            else:
                self.dependencies[i][0] -= 1
        if index == -1:
            raise MissingArtifactError()
        if self.dependencies[index][0] == len(self.element) - 1:
            self.element[self.dependencies[index][0] -1].tail = self.element[self.dependencies[index][1]].tail
        del self.element[self.dependencies[index][0]]
        del self.dependencies[index]
        self._modified = True

    def __str__(self):
        return "\n".join([ str(a) for a in self.dependencies ])

    def __iter__(self):
        return DependencyIterator(self)


class Modules(nxpy.xml.util.SequenceElement):
    """
    Map a POM's modules element to a mutable sequence.
    
    Supports addition and removal of modules by name.
    """
    def __init__(self, root):
        super(Modules, self).__init__(root, "modules", "module", _ns.url)

    def _saved(self):
        self.modified = False

    def contains(self, module):
        return module in self

    def add(self, module):
        if self.contains(module):
            raise ModuleAlreadyPresentError()
        i = 0
        for m in self:
            if m > module:
                break
            i += 1
        self.insert(i, module)

    def remove(self, module):
        i = 0
        for m in self:
            if m == module:
                break
            i += 1
        if i == len(self):
            raise MissingModuleError()
        del self[i]


class Scm(object):
    def __init__(self, element):
        self._modified = False
        self.element = element
        self._connection = _ns.find(element, "connection")
        self._developerConnection = _ns.find(element, "developerConnection")
        self._url = _ns.find(element, "url")

    connection = nxpy.xml.util.make_property("_connection")
    developerConnection = nxpy.xml.util.make_property("_developerConnection")
    url = nxpy.xml.util.make_property("_url")

    @property
    def modified(self):
        return self._modified


class Repository(object):
    def __init__(self, element):
        self._modified = False
        self.element = element
        self._id = _ns.find(element, "id")
        self._name = _ns.find(element, "name")
        self._url = _ns.find(element, "url")

    id = nxpy.xml.util.make_property("_id")
    name = nxpy.xml.util.make_property("_name")
    url = nxpy.xml.util.make_property("_url")

    @property
    def modified(self):
        return self._modified


class DistributionManagement(object):
    def __init__(self, element):
        self.element = element
        self.repository = Repository(_ns.find(element, "repository"))
        self.snapshotRepository = Repository(_ns.find(element, "snapshotRepository"))

    def _saved(self):
        self.repository._modified = False
        self.snapshotRepository._modified = False

    @property
    def modified(self):
        return self.repository.modified or self.snapshotRepository.modified


class Properties(nxpy.xml.util.MappingElement):
    def __init__(self, parent):
        super(Properties, self).__init__(parent, "properties", _ns.url)
        
    
class Pom(object):
    _root = ( r'<project xmlns="http://maven.apache.org/POM/4.0.0"'
              r' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n' 
              r'    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0' 
              r' http://maven.apache.org/maven-v4_0_0.xsd">' )

    def __init__(self, path):
        self.path = os.path.realpath(path)
        self.dir = os.path.split(self.path)[0]
        try:
            self.tree = nxpy.xml.util.parse(path)
        except IOError:
            raise BadPomFileError(path)
        self.root = self.tree.getroot()
        self.artifact = nxpy.maven.artifact.Artifact(self.root)
        self.dependencies = Dependencies(_ns.find(self.root, "dependencies"))
        parent = _ns.find(self.root, "parent")
        try:
            self.parent = nxpy.maven.artifact.Artifact(parent)
        except ( AttributeError, TypeError ):
            self.parent = None
        self.scm = None
        scm = _ns.find(self.root, "scm")
        if scm  is not None:
            self.scm = Scm(scm)
        self.modules = Modules(self.root)
        depmgmt = None
        if self.artifact.packaging == "pom":
            deps = _ns.find(self.root, "dependencyManagement")
            if deps is not None:
                depmgmt = _ns.find(deps, "dependencies")
        self.dependencyManagement = Dependencies(depmgmt)
        self.properties = Properties(self.root)
        self.assembly_descriptor = None
        build = _ns.find(self.root, "build")
        if build is not None:
            plugins = _ns.find(build, "plugins")
            if plugins is not None:
                for p in plugins.getchildren():
                    if _ns.find(p, "artifactId").text == "maven-assembly-plugin":
                        ad = _ns.find(_ns.find(p, "configuration"), "descriptors")[0]
                        self.assembly_descriptor = os.path.normpath(os.path.join(self.dir, 
                                ad.text))
                        break
        self.distributionManagement = None
        dm = _ns.find(self.root, "distributionManagement")
        if dm is not None:
            self.distributionManagement = DistributionManagement(dm)
        self._writer = nxpy.xml.util.Writer(self._root, None, 4)

    def qualified_name(self, full=False):
        return self.artifact.qualified_name(full)

    @property
    def modified(self):
        return ( self.artifact.modified or ( self.parent and self.parent.modified ) or
                 self.modules.modified or self.dependencies.modified or 
                 self.dependencyManagement.modified or ( self.scm and self.scm.modified ) or 
                 self.properties.modified or 
                 ( self.distributionManagement and self.distributionManagement.modified ) )

    def write(self, where):
        if not where:
            where = self.path
        self._writer.write(self.root, where)

    def save(self):
        if self.modified:
            self.write(None)
            self.artifact._modified = False
            if self.parent:
                self.parent._modified = False
            self.scm._modified = False
            self.dependencies._saved()
            self.dependencyManagement._saved()
            self.distributionManagement._saved()
            self.modules._saved()
            self.properties.modified = False
            return True
        return False



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Temporary files and directories.

Requires at least Python 2.6

"""

from __future__ import absolute_import

import logging
import os
import shutil
import stat
import tempfile

import nxpy.core.file_object
import nxpy.core.path
import nxpy.core.past


class TempFile(nxpy.core.file_object.WritableFileObject):
    r"""A temporary file that implements the context manager protocol.
    
    Wrap a :py:func:`tempfile.NamedTemporaryFile` generated file-like object, to ensure it is not
    deleted on close, but rather when the underlying context is closed.
    
    """
    def __init__(self, *args, **kwargs):
        r"""Create a temporary file with the given arguments."""
        # Python 2.5 compatible syntax. Doesn't work, but ensures that tests are skipped rather
        # than broken.
        # file = tempfile.NamedTemporaryFile(*args, delete=False, **kwargs)
        kwargs["delete"] = False
        if nxpy.core.past.V_2_7.at_most() and "encoding" in kwargs:
            del kwargs["encoding"]
        file_ = tempfile.NamedTemporaryFile(*args, **kwargs)
        super(TempFile, self).__init__(file_)
    
    @property
    def name(self):
        r"""Return the actual file name."""
        return self._file.name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._file.close()
        os.remove(self._file.name)
        return False


class TempDir(object):
    r"""A temporary directory that implements the context manager protocol.
    
    The directory is removed when the context is exited from. Uses :py:func:`tempfile.mkdtemp` to 
    create the actual directory.
    
    """
    def __init__(self, *args, **kwargs):
        r"""Create a temporary directory with the given arguments."""
        self.dir = tempfile.mkdtemp(*args, **kwargs)
        mode = os.stat(self.dir).st_mode
        os.chmod(self.dir, mode | stat.S_IWRITE)

    @property
    def name(self):
        r"""Return the directory name."""
        return self.dir

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        nxpy.core.path.blasttree(self.dir)
        return False



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Utility functions that deal with non-string sequences.

"""

from __future__ import absolute_import

import six


def make_tuple(arg):
    r"""An alternate way of creating tuples from a single argument.
    
    A single string argument is turned into a single element tuple and a dictionary argument is 
    turned into a tuple of its items. Otherwise it works like the standard tuple constructor.
    
    """
    try:
        if isinstance(arg, six.string_types):
            return ( arg, )
        elif isinstance(arg, dict):
            return tuple(arg.items())
        else:
            return tuple(arg)
    except TypeError:
        return ( arg, )



# nxpy.ply package -----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Class wrapper for PLY's lex module.

"""

from __future__ import absolute_import

import re

import ply.lex


class Scanner(object):
    def __init__(self, debug=False, ignorecase=False):
        self.debug = debug
        reflags = 0
        if ignorecase:
            reflags = re.IGNORECASE 
        self.lexer = ply.lex.lex(module=self, debug=self.debug, reflags=reflags)

    def reset(self, input_):
        self.lexer.input(input_)
    
    def token(self):
        return self.lexer.token()



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2013 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Identification and enforcement of supported Python releases.

"""

from __future__ import absolute_import

import sys


class Version(object):
    r"""Identifies a Python release in a way that is convenient for comparison and printing."""
    
    def __init__(self, version):
        self.version = version

    def at_least(self):
        r"""Return *True* if the current Python version is equal or higher than *self*."""
        return sys.hexversion >= self.version
    
    def at_most(self):
        r"""Return *True* if the current Python version is equal or lower than *self*."""
        return sys.hexversion < ( self.version | 0xff00 )
    
    def __str__(self):
        v = hex(self.version)
        return v[2] + "." + v[4] + "." + v[6]

V_3_6 = Version(0x030600f0)

V_3_5 = Version(0x030500f0)

V_3_4 = Version(0x030400f0)

V_3_3 = Version(0x030300f0)

V_3_2 = Version(0x030200f0)

V_2_7 = Version(0x020700f0)

V_2_6 = Version(0x020600f0)

V_2_5 = Version(0x020500f0)

def enforce_at_least(version):
    r"""Assert that the current Python version is equal or higher than *version*."""
    assert version.at_least(), "Requires at least Python " + str(version)
 
def enforce_at_most(version):
    r"""Assert that the current Python version is equal or lower than *version*."""
    assert version.at_most(), "Requires at most Python " + str(version)
        


# nxpy.msvs package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Manipulation of Microsoft Visual Studio solution files.

Currently supports Visual Studio 2008 (9.0) only.

Requires at least Python 2.6 and `ply <https://pypi.python.org/pypi/ply>`_.

"""

# Alllows for more gracious failure on Python 2.5
from __future__ import with_statement
from __future__ import absolute_import

import os.path

import six

import nxpy.core.file

import nxpy.msvs._impl.solution.parser
import nxpy.msvs._impl.solution.scanner


class Description(object):
    def __init__(self, ast, disp):
        self.pos = ast.pos + disp
        self.len = ast.len
        self._value = ast.value
        self.modified = False
        
    def _get_value(self):
        return self._value
    def _set_value(self, value):
        if value != self._value:
            self._value = value
            self.modified = True
    value = property(_get_value, _set_value)

    def update(self, text):
        if len(self.value) != 0:
            t = [ text[:self.pos] ]
            if self.len == 0:
                t.append("\t")
            v = "Description = " + self.value
            t.append(v)
            if self.len == 0:
                t.append("\n")
            t.append(text[self.pos+self.len:])
            self.len = len(v)
            return "".join(t)
        else:
            t = [ text[:self.pos].rstrip() ]
            t.append(text[self.pos+self.len].lstrip(" \t"))
            self.pos = len(t[0])
            self.len = 0
            return "".join(t)
        
            
class Solution(object):
    def __init__(self, path):
        self.path = os.path.realpath(path)
        self.dir, self.file = os.path.split(self.path)
        self.name = os.path.splitext(self.file)[0]
        ast = self._read(debug=1)
        self.text = ast.text
        self.version = ast.version
        self._description = Description(ast.description, len(self.text.splitlines(True)[0]))
        self.projects = {}
        for p in ast.projects:
#            l = [ self.dir, ]
#            l.extend(p.path.split("\\"))
#            self.projects[p.name] = os.path.join(*l)
            self.projects[p.name] = os.path.join(self.dir, p.path)

    def _read(self, debug=False):
        scanner = nxpy.msvs._impl.solution.scanner.Scanner(debug=debug)
        parser = nxpy.msvs._impl.solution.parser.Parser(scanner, debug=debug)
        with nxpy.core.file.open_(self.path, "r", encoding="utf-8") as f:
            return parser.parse(f.read())

    def _get_description(self):
        return self._description.value
    def _set_description(self, value):
        self._description.value = value
    description = property(_get_description, _set_description)

    def write(self, where):
        if not where:
            where = self.path            
        if isinstance(where, six.string_types):
            f = nxpy.core.file.open_(where, "w+", encoding="utf-8")
        else:
            f = where
        try:
            t = self._description.update(self.text)
            if t is None:
                t = self.text
            f.write(t)
            self.text = t
        finally:
            f.close()
        
    def save(self):
        if self._description.modified:
            self.write(None)
            return True
        return False

    def __str__(self):
        return "\n\n".join([ " ".join([ "Solution", self.name, self.version, self.dir ]) ] + 
                            [ str(p) for p in self.projects.values() ])



# nxpy.svn package -----------------------------------------------------------

# Copyright Nicola Musatti 2011 - 2015
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Working copy manipulation.

"""

from __future__ import absolute_import

import os.path

import nxpy.command.command
import nxpy.svn.svn
import nxpy.svn.url


class ModifiedError(Exception):
    r"""Raised when attempting to tag or branch a working copy that contains changes."""


class NotOnBranchError(Exception):
    r"""Raised when attempting to delete a working copy that is not on the requested branch."""


class NotOnTagError(Exception):
    r"""Raised when attempting to delete a working copy that is not on the requested tag."""


class Wcopy(object):
    r"""
    A working copy obtained by checking out a *Url*.

    """
    def __init__(self, dir_, url=None, username="", password=""):
        r"""
        Initialize attributes.
        
        If *url* is not None, perform a checkout, otherwise check that *dir\_* points to a valid
        working copy.
        
        """
        self.dir = dir_
        self.username = username
        self.password = password
        self.svn = nxpy.svn.svn.Svn()
        if url:
            self.svn.checkout(url, self.dir, ignore_externals=False)
            self.url = url
        else:
            self.url = nxpy.svn.url.Url(self.svn.info(self.dir).url)

    def __str__(self):
        return self.dir

    def commit(self):
        self.svn.commit(self.dir, username=self.username, password=self.password)

    def update(self, ignore_externals=False):
        self.svn.update(self.dir, ignore_externals=ignore_externals)

    # If 'url' exists do nothing, otherwise copy the working copy
    def _copy(self, url):
        info = None
        try:
            info = self.svn.info(url)
        except nxpy.command.command.Error:
            pass
        if not info:
            if self.svn.status(self.dir):
                raise ModifiedError(self.dir)
            self.svn.copy(self.dir, url, username=self.username, password=self.password)
        
    def branch(self, label):
        if self.url.isbranch(label):
            return None
        else:
            u = self.url.getbranch(label)
            self._copy(u)
            return u

    def delete_branch(self, label):
        if not self.url.isbranch(label):
            raise NotOnBranchError(self.dir)
        self.svn.delete(self.url, username=self.username, password=self.password)
        self.url = None
        
    def tag(self, label):
        if self.url.istag(label):
            return None
        else:
            u = self.url.gettag(label)
            self._copy(u)
            return u

    def delete_tag(self, label):
        if not self.url.istag(label):
            raise NotOnTagError(self.dir)
        self.svn.delete(self.url, username=self.username, password=self.password)
        self.url = None

    def delete_path(self, path, keep_local=False):
        path = os.path.join(self.dir, path)
        self.svn.delete(path, keep_local=keep_local)

    def getexternals(self):
        return self.svn.getexternals(self.dir)
    
    def setexternals(self, ext):
        self.svn.setexternals(ext, self.dir, username=self.username, password=self.password)
        self.update(False)
        self.commit()
        self.update(False)

    def getignore(self):
        return self.svn.getignore(self.dir)
    
    def setignore(self, ign):
        self.svn.setignore(ign, self.dir, username=self.username, password=self.password)
        self.update(False)
        self.commit()
        self.update(False)



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Interactive program driver.

"""

from __future__ import absolute_import

import os
import re
import subprocess
import sys
import time

import six

import nxpy.core.async_subprocess
import nxpy.command.error


# Output prefixes

OUTPUT  = "OUT> "
ERROR   = "ERR> "
EXCEPT  = "EXC> "
COMMAND = "CMD> "


# Type of output

FILE = 0
STRING = 1


def _format(input_, output=STRING, type_=OUTPUT):
    r"""
    Prepends a prefix to each *input_* line as specified by *type_* and copies it to *output*.
    Returns *output*.

    """
    if type_ not in ( OUTPUT, ERROR, EXCEPT, COMMAND ):
        raise nxpy.command.error.BadLogFormat(str(type_) + ": Unknown format")
    if isinstance(input_, six.string_types):
        if input_ and input_[-1] != '\n':
            input_ = input_ + '\n'
        input_ = six.StringIO(input_)
    if output in ( FILE, STRING ):
        out = six.StringIO()
    else:
        out = output
    for line in input_:
        out.write(type_)
        out.write(line)
    if output == STRING:
        return out.getvalue()
    return out


class BadCommand(Exception):
    r"""Raised on a command execution failure"""
    
    def __init__(self, cmd, err):
        r"""Takes the failed command and the contents of the error stream."""
        self.command = cmd
        self.stderr = err
        msg = _format(cmd, FILE, COMMAND)
        if err:
            msg = _format(err, msg, ERROR)
        super(BadCommand, self).__init__("".join(msg))


EXP_OUT = 0
EXP_ERR = 1

def waitOutput(out, err):
    r"""Wait for any output."""
    return out

def waitError(out, err):
    r"""Wait for any error."""
    return err


class LineWaiter(object):
    r"""Wait for *count* lines of output."""
    def __init__(self, count):
        self.count = count
        self.n = 0
        
    def __call__(self, out, err):
        self.n += out.count('\n')
        return self.n >= self.count


class StringWaiter(object):
    r"""Wait for a specific *string* in the *where* stream."""
    def __init__(self, string, where):
        self.string = string
        self.where = where
    
    def __call__(self, out, err):
        if self.where == EXP_OUT:
            o = out
        else:
            o = err
        return o.find(self.string) != -1


class RegexpWaiter(object):
    r"""Wait for a match to a given *regexp*, passed either compiled or as a string."""
    def __init__(self, regexp, where):
        if isinstance(regexp, six.string_types):
            self.regexp = re.compile(regexp, re.MULTILINE)
        else:
            self.regexp = regexp
        self.where = where
    
    def __call__(self, out, err):
        if self.where == EXP_OUT:
            o = out
        else:
            o = err
        return self.regexp.search(o)


class Timer(object):
    r"""
    A collaborative timer class.
    Support a polling mechanism by keeping track of the amount of time to wait before the next
    attempt, according to different policies.
    
    """    
    def __init__(self, timeout=0, retries=0, interval=0.1, quantum=0.01):
        """
        Specify an overall *timeout*, a number of *retries* and/or an *interval* between them.
        The next attempt will not take place before a *quantum* has passed. Timings are expressed
        in seconds. If a timeout is specified it will take precedence over the other arguments; in
        that case the number of retries will take precedence over the interval. If neither a timeout
        nor a number of retries are specified the overall timer will never expire.
        
        """
        if retries < 0:
            raise nxpy.command.error.TimerError("retries must be equal or greater than 0")
        elif retries == 0:
            self.retries = -1
        else:
            self.retries = retries
        self.count = 0
        self.interval = interval
        if timeout > 0:
            self.timeout = timeout
            self.end = time.time() + timeout
            if retries > 1:
                self.interval = timeout / retries
        else:
            self.timeout = 0
            self.end = 0
        if quantum < 0:
            raise nxpy.command.error.TimerError("quantum must be equal or greater than 0")
        self.quantum = quantum

    def getInterval(self):
        r"""
        Return the next wait interval.
        Call after each attempt in order to know how long to wait for.
        
        """
        if self.end and self.retries > 1:
            self.interval = max((self.end-time.time())/self.retries, self.quantum)
        return self.interval

    def expired(self):
        r"""
        Indicate whether the current timer expired.
        Use as polling loop control condition.
        
        """
        self.count += 1
        return self.timeout and self.end < time.time() or ( self.retries - self.count == 0 )

    def reset(self):
        r"""Reset the timer."""
        self.count = 0
        if self.timeout:
            self.end = time.time() + self.timeout


class BaseInterpreter(object):
    r"""
    Controls the execution of an interactive program in a sub-process.
    Provides means to send input to the controlled process and to check different conditions on its
    output and error streams.
    
    """
    def __init__(self, popen):
        r"""
        Creates an interpreter instance. *popen* is a :py:class:`.Popen`-like object which must
        support non-blocking I/O.
        
        """
        self.log = False
        self.popen = popen

    def setLog(self, log):
        r"""
        If *log* is *True*, enable logging of command output and error, otherwise disable it.
        
        """
        self.log = log
        
    def _log(self, log):
        r"""
        Check whether logging should be enabled. Usually *log* is passed from the calling method.
        
        """
        if log != None:
            return log
        else:
            return self.log

    def send_cmd(self, cmd, log=None):
        r"""
        Write *cmd* to the interpreter's input, optianally logging it. If *log* is not *None*,
        override the global setting.
        
        """
        try:
            if self._log(log):
                _format(cmd, sys.stderr, COMMAND)
            self.popen.send(cmd + os.linesep)
        except Exception:
            e = sys.exc_info()[1]
            raise BadCommand(cmd, str(e.args))
            
    def expect_any(self, **kwargs):
        r"""Expect any output."""
        return self.expect(waitOutput, **kwargs)

    def expect_lines(self, count=1, **kwargs):
        r"""Expect *count* lines of output."""
        return self.expect(LineWaiter(count), **kwargs)
    
    def expect_string(self, string, where=EXP_OUT, **kwargs):
        r"""Expect a *string* in the *where* stream."""
        return self.expect(StringWaiter(string, where), **kwargs)  

    def expect_regexp(self, regexp, where=EXP_OUT, **kwargs):
        r"""
        Expect to find a match for the *regexp* regular expression within the *where* stream.
        
        """
        return self.expect(RegexpWaiter(regexp, where), **kwargs)  

    def expect(self, cond=None, timeout=0, retries=0, interval=0.01, 
            quantum=0.01, raise_on_error=True, log=None):
        r"""
        Express expectations on the outcome of a command.
        
        *cond* is a two argument callable which will be passed the command's standard output and
        standard error, and which should return *True* if the expectation is satisfied. For the
        other arguments see the documentation for the :py:class:`.Timer` class.
        
        """
        try:
            out_list = []
            err_list = []
            timer = Timer(timeout, retries, interval, quantum)
            while not timer.expired():
                out = self.popen.recv()
                if out:
                    out_list.append(out)
                err = self.popen.recv_err()
                if err:
                    err_list.append(err)
                if cond and cond(out, err):
                    break
                if out or err:
                    timer.reset()
                    t = timer.quantum
                else:
                    t = timer.getInterval()
                if self._log(log):
                    sys.stderr.write("END: %s  SLEEP: %f  SIZE: %d\n" % 
                            (time.ctime(timer.end), t, len(out) + len(err)))
                if t > 0:
                    time.sleep(t)
            else:
                if cond:
                    raise nxpy.command.error.TimeoutError(err)
        finally:
            out = ''.join(out_list)
            err = ''.join(err_list)
            if self._log(log):
                _format(out, sys.stderr)
                _format(err, sys.stderr, ERROR)
        if raise_on_error and err:
            raise nxpy.command.error.ExpectError(err)
        return out, err

    def run(self, cmd, log=None, **kwargs):
        r"""Executes the command and waits for the expected outcome or an error."""
        self.send_cmd(cmd, log=log)
        kwargs['log'] = log
        try:
            return self.expect(**kwargs)
        except nxpy.command.error.ExpectError:
            e = sys.exc_info()[1]
            raise BadCommand(cmd, e.args[0])


class Interpreter(BaseInterpreter):
    r"""
    The actual Interpreter class.
    
    This implementation uses a :py:class:`.core.async_subprocess.AsyncPopen` instance.
    
    """
    def __init__(self, cmd):
        super(Interpreter, self).__init__(nxpy.core.async_subprocess.AsyncPopen(cmd.split(), 
                            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE))



# nxpy.scons package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Microsoft Visual Studio configuration.

"""

from __future__ import absolute_import

import SCons.Builder

import nxpy.scons.clean_action
import nxpy.scons.project


def _clean_project(env, target):
    env.CleanAction(target, SCons.Action(env.subst('msbuild "$PROJECT" /t:Clean', 
            target=target)))

    
def environment(version=None, 
        script='C:/Program Files (x86)/Microsoft Visual Studio 9.0/Common7/Tools/vsvars32.bat'):
    bldr = SCons.Builder(action = 'msbuild "$PROJECT"')
    proj = nxpy.scons.project.Project()
    kwargs = { 'BUILDERS' : {'Assembly' : bldr}, 'PROJECT' : proj }
    if version:
        kwargs['MSVC_VERSION'] = version
    elif script:
        kwargs['MSVC_USE_SCRIPT'] = script
    env = SCons.Environment(**kwargs)
    env.AddMethod(nxpy.scons.clean_action.clean_action, 'CleanAction')
    env.AddMethod(proj.add, 'AddProject')
    env.AddMethod(_clean_project, 'CleanProject')
    return env



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2015
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Allow asynchronous interaction with a subprocess.

This module was taken from `this recipe 
<http://code.activestate.com/recipes/440554-module-to-allow-asynchronous-subprocess-use-on-win/>`_ 
in the `ActiveState Code Recipes website 
<http://code.activestate.com/recipes/langs/python/>`_, with only minor
modifications. This is the original description: ::

  Title:        Module to allow Asynchronous subprocess use on Windows and Posix platforms
  Submitter:    Josiah Carlson (other recipes)
  Last Updated: 2006/12/01
  Version no:   1.9
  Category:     System 

On Windows `pywin32 <https://pypi.python.org/pypi/pywin32>`_ is required.

"""

from __future__ import absolute_import

import os
import subprocess
import errno
import locale
import time
import sys

import six

import nxpy.core.past

if nxpy.core.past.V_2_5.at_most():
    def bytearray(s, e):
        return s
    
## PIPE = subprocess.PIPE

mswindows = (sys.platform == "win32")

if mswindows:
    from win32file import ReadFile, WriteFile
    from win32pipe import PeekNamedPipe
    import msvcrt
    import pywintypes
    
else:
    import select
    import fcntl

class AsyncPopen(subprocess.Popen):
    r"""
    An asynchronous variant to :py:class:`subprocess.Popen`, which doesn't block on incomplete I/O 
    operations.
    
    Note that the terms input, output and error refer to the controlled program streams,
    so we receive from output or error and we send to input.
    
    """

    def __init__(self, cmd, encoding=None, **kwargs):
        r"""
        Execute *cmd* in a subprocess, using *encoding* to convert to and from binary data written
        or read from/to the subprocess's input, output and error streams.

        Additional keyword arguments are as specified by :py:func:`subprocess.Popen.__init__`
        method.

        """
        self._encoding = encoding
        if self._encoding is None:
            self._encoding = locale.getpreferredencoding()
        super(AsyncPopen, self).__init__(cmd, **kwargs)

    def recv(self, maxsize=None):
        r"""Receive at most *maxsize* bytes from the subprocess's standard output."""
        return self._recv('stdout', maxsize)
    
    def recv_err(self, maxsize=None):
        r"""Receive at most *maxsize* bytes from the subprocess's standard error."""
        return self._recv('stderr', maxsize)

    def send_recv(self, input_='', maxsize=None):
        r"""
        Send *input_* to the subprocess's standard input and then receive at most *maxsize* bytes
        from both its standard output and standard error.

        """
        return self.send(input_), self.recv(maxsize), self.recv_err(maxsize)

    def get_conn_maxsize(self, which, maxsize):
        r"""
        Return *which* output pipe (either stdout or stderr) and *maxsize* constrained to the 
        [1, 1024] interval in a tuple.
        
        """
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return getattr(self, which), maxsize
    
    def _close(self, which):
        getattr(self, which).close()
        setattr(self, which, None)
    
    if mswindows:
        def send(self, input_):
            r"""Send *input_* to the subprocess's standard input."""
            if not self.stdin:
                return None

            try:
                x = msvcrt.get_osfhandle(self.stdin.fileno())
                (errCode, written) = WriteFile(x, bytearray(input_, self._encoding))
            except ValueError:
                return self._close('stdin')
#            except (subprocess.pywintypes.error, Exception):
            except pywintypes.error:
                why = sys.exc_info()[1]
                if why.winerror in (109, errno.ESHUTDOWN):
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None
            
            try:
                x = msvcrt.get_osfhandle(conn.fileno())
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
                if maxsize < nAvail:
                    nAvail = maxsize
                if nAvail > 0:
                    (errCode, read) = ReadFile(x, nAvail, None)
            except ValueError:
                return self._close(which)
#            except (subprocess.pywintypes.error, Exception):
            except pywintypes.error:
                why = sys.exc_info()[1]
                if why.winerror in (109, errno.ESHUTDOWN):
                    return self._close(which)
                raise
            
            if self.universal_newlines:
                read = self._translate_newlines(read, self._encoding)
            else:
                read = read.decode(self._encoding)
            return read

    else:
        def send(self, input_):
            r"""Send *input_* to the subprocess's standard input."""
            if not self.stdin:
                return None

            if not select.select([], [self.stdin], [], 0)[1]:
                return 0

            try:
                written = os.write(self.stdin.fileno(), bytearray(input_, self._encoding))
            except OSError:
                why = sys.exc_info()[1]
                if why[0] == errno.EPIPE: #broken pipe
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None
            
            flags = fcntl.fcntl(conn, fcntl.F_GETFL)
            if not conn.closed:
                fcntl.fcntl(conn, fcntl.F_SETFL, flags| os.O_NONBLOCK)
            
            try:
                if not select.select([conn], [], [], 0)[0]:
                    return ''
                
                r = conn.read(maxsize)
                if not r:
                    return self._close(which)
    
                if self.universal_newlines:
                    r = self._translate_newlines(r)
                else:
                    r = r.decode(self._encoding)
                return r
            finally:
                if not conn.closed:
                    fcntl.fcntl(conn, fcntl.F_SETFL, flags)

message = "Other end disconnected!"

def recv_some(p, t=.1, e=1, tr=5, stderr=0):
    r"""
    Try and receive data from :py:class:`.AsyncPopen` object *p*'s stdout in at most *tr* tries and
    with a timeout of *t*. If *stderr* is True receive from the subprocess's stderr instead.
    
    """
    if tr < 1:
        tr = 1
    x = time.time()+t
    y = []
    r = six.b('')
    pr = p.recv
    if stderr:
        pr = p.recv_err
    while time.time() < x or r:
        r = pr()
        if r is None:
            if e:
                raise Exception(message)
            else:
                break
        elif r:
            y.append(r)
        else:
            time.sleep(max((x-time.time())/tr, 0))
    return ''.join(y)
    
def send_all(p, data):
    r"""Send all of *data* to :py:class:`.AsyncPopen` object *p*'s stdin."""
    while len(data):
        sent = p.send(data)
        if sent is None:
            raise Exception(message)
        data = data[sent:-1]



# nxpy.msvs package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
A C# or Visual Basic.NET source file that conventionally contains information about its containing 
assembly, such as its version number.

"""

from __future__ import absolute_import

import encodings.utf_8_sig
import logging
import os.path
import re

import six

import nxpy.core.backup_file
import nxpy.core.file

import nxpy.core._impl.log


_log = logging.getLogger(__name__)


class BadFileExtensionError(Exception):
    r"Raised when an unknown file extension is found"


class _LineRep(object):
    def __init__(self, start, ai):
        self._start = start
        self._ai = ai
        self.found = -1

    def __call__(self, match):
        _log.debug("ai: " + str(match.start()) + " cm: " + str(self._start))
        if self._start > -1 and match.start() > self._start:
            return match.group(0)
        if match.group(3) == "AssemblyVersion":
            s = self._ai._version
            self.found = 0
        elif match.group(3) == "AssemblyFileVersion":
            s = self._ai._file_version
            self.found = 1
        elif match.group(3) == "AssemblyInformationalVersion":
            s = self._ai._product_version
            self.found = 2
        else:
            return match.group(0)
        return ( match.group(1) + match.group(2) + match.group(3) + "(\"" + s + "\")" + 
                match.group(5) )

        
class AssemblyInfo(object):
    _ai_re = re.compile(r"([<[])([Aa]ssembly:\s+)(\w+)\(\"([^\"]+)\"\)([>\]])")
    _cm_re = re.compile(r"(?:')|(?://)|(/\*)")
    
    def __init__(self, path):
        self.path = path
        self._ext = os.path.splitext(path)[1].lower()
        self._version = None
        self._file_version = None
        self._product_version = None
        self._modified = False 
        comment = False
        ebc = -1
        for l in nxpy.core.file.open_(path, "r", encoding="utf-8"):
            if comment:
                ebc = l.find("*/")
            if not comment or ebc != -1:
                ai_match = self._ai_re.search(l)
                cm_match = self._cm_re.search(l)
                if ai_match and ( ( comment and ebc < ai_match.start() ) or ( not cm_match or 
                        ai_match.start() < cm_match.start() ) ):
                    _log.debug("ai: " + str(ai_match.start()) + " cm: " + 
                            str(cm_match.start() if cm_match else -1))
                    if ai_match.group(3) == "AssemblyVersion":
                        self._version = ai_match.group(4)
                    elif ai_match.group(3) == "AssemblyFileVersion":
                        self._file_version = ai_match.group(4)
                    elif ai_match.group(3) == "AssemblyInformationalVersion":
                        self._product_version = ai_match.group(4)
                if cm_match and cm_match.group(1) and ebc < cm_match.start(1):
                    comment = True
                elif ebc != -1:
                    comment = False
                    ebc = -1

    def _get_version(self):
        return self._version
    def _set_version(self, v):
        self._version = v
        self._modified = True
    version = property(_get_version, _set_version)

    def _get_file_version(self):
        return self._file_version
    def _set_file_version(self, v):
        self._file_version = v
        self._modified = True
    file_version = property(_get_file_version, _set_file_version)

    def _get_product_version(self):
        return self._product_version
    def _set_product_version(self, v):
        self._product_version = v
        self._modified = True
    product_version = property(_get_product_version, _set_product_version)

    def _attr_fmt(self):
        if self._ext == ".vb":
            return r'<Assembly: %s("%s")>'
        elif self._ext == ".cs":
            return r'[assembly: %s("%s")]'
        else:
            raise BadFileExtensionError(self._ext)

    def write(self, dest):
        src = None
        bck = None
        if not dest:
            dest = self.path
            bck = nxpy.core.backup_file.BackupFile(self.path, 
                    mode=nxpy.core.backup_file.BackupFile.MOVE)
            bck.save()
            bck.open(mode=nxpy.core.backup_file.BackupFile.BINARY)
            src = encodings.utf_8_sig.StreamReader(bck)
        else:
            src = encodings.utf_8_sig.StreamReader(open(self.path, "rb"))
        if isinstance(dest, six.string_types):
            dest = encodings.utf_8_sig.StreamWriter(open(dest, "wb+"))
        else:
            dest = encodings.utf_8_sig.StreamWriter(dest)
            dest.seek(0)
        try:
            comment = False
            ebc = -1
            found = [ False, False, False ]
            for l in src:
                if comment:
                    ebc = l.find("*/")
                if not comment or ebc != -1:
                    cm_match = self._cm_re.search(l)
                    lr = _LineRep(cm_match.start() if cm_match else -1, self)
                    l = self._ai_re.sub(lr, l)
                    if lr.found != -1:
                        found[lr.found] = True
                    if cm_match and cm_match.group(1) and ebc < cm_match.start(1):
                        comment = True
                    elif ebc != -1:
                            comment = False
                            ebc = -1
                dest.write(l)
            if not found[0]:
                dest.write(self._attr_fmt() % ( "AssemblyVersion", self.version ))
            if not found[1]:
                dest.write(self._attr_fmt() % ( "AssemblyFileVersion", self.file_version ))
            if not found[2]:
                dest.write(self._attr_fmt() % ( "AssemblyInformationalVersion", 
                        self.product_version ))
            if bck:
                bck.commit()
        except:
            if bck:
                bck.rollback()
            raise
        finally:
            dest.close()

    def save(self):
        if self._modified:
            self.write(None)
            return True
        return False



# nxpy.ccase package ---------------------------------------------------------

# Copyright Nicola Musatti 2012 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Clearcase test environment definition.

"""

from __future__ import absolute_import

import os.path
import platform

from six.moves import configparser

import nxpy.test.env


class ConfigurationError(Exception):
    r"""Raised when the ClearCase test environment is not properly set."""


class Env(nxpy.test.env.EnvBase):
    r"""Environment configuration information for ClearCase related tests."""
    
    src_link_name = "nxpy_test_link"
    r"""Name for a link created when testing the *ln()* method."""
    
    backup_ext = ".test"
    r"""Extension used for backup files."""

    def __init__(self):
        r"""Takes initialization information from a conventionally placed configuration file."""
        
        nxpy.test.env.EnvBase.__init__(self, "conf")
        self.parser = configparser.SafeConfigParser()
        self.parser.read(os.path.join(self.elem_dir, "ccase.ini"))
        self.section = platform.node()
        if not self.parser.has_section(self.section):
            self.section = "localhost"
            if not self.parser.has_section(self.section):
                raise ConfigurationError("ccase.ini has no section for this computer")

        self.src_view_dir = self._get("src_view_dir")
        self.src_view_tag = self._get("src_view_tag")
        self.src_stream_tag = self._get("src_stream_tag")
        self.src_stream_sel = self._get("src_stream_sel")
        self.src_proj_sel = self._get("src_proj_sel")
        self.relative_src_dir = self._get("relative_src_dir")
        self.src_file_name = self._get("src_file_name")
        self.dest_view_dir = self._get("dest_view_dir")
        self.dest_view_tag = self._get("dest_view_tag")
        self.dest_stream_sel = self._get("dest_stream_sel")
        self.proj_vob = self._get("proj_vob")
        self.lsview_tags = self._get("lsview_tags")
        self.activity = self._get("activity")
        self.test = self._getboolean("test")
        
        self.src_dir_path = os.path.join(self.src_view_dir, self.relative_src_dir)
        self.src_file_path = os.path.join(self.src_dir_path, self.src_file_name)
        self.backup_file_name = self.src_file_name + self.backup_ext
        self.src_stream_pattern = "\*" + self.src_stream_tag + "\*"

    def _get(self, option):
        return self.parser.get(self.section, option)

    def _getboolean(self, option):
        return self.parser.getboolean(self.section, option)


def get_env(test):
    r"""
    If the environment is correctly configured an instance of *Env* is returned, otherwise
    the current test is skipped. *test* is a :py:class:`unittest.TestCase` instance containing the
    test currently being executed.
    
    """
    try:
        return Env()
    except nxpy.test.env.TestEnvNotSetError:
        test.skipTest("Test environment not set")
        return None



# nxpy.etree package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
ElemenTree related utility classes and functions.

Requires at least Python 2.6. Simple import breaks on Python 2.5

"""
from __future__ import absolute_import

import collections
import re
import xml.etree.ElementTree

import six

import nxpy.core.error
import nxpy.core.past

nxpy.core.past.enforce_at_least(nxpy.core.past.V_2_6)


def make_property(elem, key=None):
    r"""
    Creates a property on the text of element 'elem' or, if the 'key' argument is given, on its
    'key' attribute.
    
    """
    if key:
        def _get(self):
            return getattr(self, elem).get(key)
        def _set(self, value):
            getattr(self, elem).set(key, value)
            self._modified = True
        return property(_get, _set)
    else:
        def _get(self):
            return getattr(self, elem).text
        def _set(self, value):
            getattr(self, elem).text = value
            self._modified = True
        return property(_get, _set)


class QName(object):
    r"""Represents a qualified name"""

    _re = re.compile(r"\{(.*)\}(.*)")
    
    def __init__(self, tag):
        m = QName._re.match(tag)
        self.url = m.group(1)
        self.tag = m.group(2)
        
    @property
    def text(self):
        t = []
        if len(self.url) != 0:
            t.append("{{{0}}}".format(self.url))
        t.append(self.tag)
        return "".join(t)
    
    def __str__(self):
        return self.text()
    
        
class Namespace(object):
    r"""
    Represents an XML namespace and provides several utility functions that help handle a
    document without namespace tags.
    
    """
    def __init__(self, url="", element=None):
        if len(url) > 0 and element is not None:
            raise nxpy.core.error.ArgumentError(
                    "Only one between url and element should be specified")
        if element is not None:
            url = QName(element.tag).url
        self.url = url
        self.nspace = "{" + url + "}" if len(url) != 0 else ""

    def find(self, element, tag):
        return element.find(self.nspace + tag)

    def findall(self, element, tag):
        return element.findall(self.nspace + tag)

    def findtext(self, element, tag, default=None):
        return element.findtext(self.nspace + tag, default)
    
    def get_tag(self, element):
        return element.tag[len(self.nspace):]
    
    def Element(self, tag, attrib={}, **extra):
        return xml.etree.ElementTree.Element(self.nspace + tag, attrib, **extra)
    
    def SubElement(self, parent, tag, attrib={}, **extra):
        return xml.etree.ElementTree.SubElement(parent, self.nspace + tag, attrib, **extra)


class ContainerElementMixin(Namespace):
    def __init__(self, parent, root_tag, namespace=""):
        super(ContainerElementMixin, self).__init__(namespace)
        self.parent = parent
        self.root_tag = root_tag
        self.root = self.find(self.parent, self.root_tag)
        self.modified = False

    def __len__(self):
        if self.root is None:
            return 0
        return len(self.root)


class MappingElementIterator(collections.Iterator):
    def __init__(self, element):
        self.element = element
        self.iter = element.getchildren().iter()
        
    def next(self):
        return self.element.get_tag(next(self.iter))

        
class MappingElement(ContainerElementMixin, collections.MutableMapping):
    def __init__(self, parent, root_tag, namespace=""):
        ContainerElementMixin.__init__(self, parent, root_tag, namespace)

    def __getitem__(self, key):
        if self.root is None:
            raise KeyError()
        elem = self.find(self.root, key)
        if elem is None:
            raise KeyError()
        return elem.text
    
    def __setitem__(self, key, value):
        if self.root is None:
            self.root = self.SubElement(self.parent, self.root_tag)
        elem = self.find(self.root, key)
        if elem is None:
            elem = self.SubElement(self.root, key)
        self.modified = True
        elem.text = value

    def __delitem__(self, key):
        if self.root is None:
            raise KeyError()
        elem = self.find(self.root, key)
        if elem is None:
            raise KeyError()
        self.modified = True
        self.root.remove(elem)

    def __iter__(self):
        return MappingElementIterator(self)


class SequenceElement(ContainerElementMixin, collections.MutableSequence):
    def __init__(self, parent, root_tag, element_tag, namespace="", indent="    "):
        ContainerElementMixin.__init__(self, parent, root_tag, namespace)
        self.element_tag = element_tag
        self.indent = indent

    def __getitem__(self, index):
        if self.root is None:
            raise IndexError()
        return self.root[index].text
    
    def __setitem__(self, index, value):
        if self.root is None:
            self.root = self.SubElement(self.parent, self.root_tag)
        elem = None
        try:
            elem = self.root[index]
        except IndexError:
            elem = self.SubElement(self.root, self.element_tag)
        elem.text = value
        self.modified = True
    
    def __delitem__(self, index):
        if self.root is None:
            raise IndexError()
        del self.root[index]
        self.modified = True

    def insert(self, index, value):
        if self.root is None:
            self.root = self.SubElement(self.parent, self.root_tag)
        elem = self.Element(self.element_tag)
        elem.text = value
        elem.tail = self.root.tail + self.indent
        self.root.insert(index, elem)
        self.modified = True


class Writer(object):
    _name_re = re.compile(r"<([^\s]+)")
    _tag_re = re.compile(r"(</?)[^:]+:((:?[^>]+>)|(:?[^/]+/>))")

    def __init__(self, root_tag, attributes=None, tab_size=0):
        self.root_tag = root_tag
        self.tab_size = tab_size
        self.attributes = attributes
        self.name = self._name_re.search(self.root_tag).group(1)
        self._root_re = re.compile(r"(<" + self.name + r"[^>]+>)")

    def marshal(self, node):
        s = None
        if nxpy.core.past.V_2_7.at_most():
                s = xml.etree.ElementTree.tostring(node)
        else:
                s = xml.etree.ElementTree.tostring(node, encoding="unicode")
        s = self._tag_re.sub(r"\1\2", s)
        s = self._root_re.sub(self.root_tag, s, 1)
        if self.tab_size > 0:
            s = s.replace("\t", " " * self.tab_size)
        if self.attributes is not None:
            d = ( '<?xml version="' + self.attributes.get("version", "1.0") + 
                    '" encoding="' + self.attributes.get("encoding", "UTF-8") + '"')
            if "standalone" in self.attributes:
                d += ' standalone="' + self.attributes["standalone"] + '"'
            d += "?>\n"
            s = d + s
        return s + "\n\n"

    def write(self, node, where):
        if isinstance(where, six.string_types):
            f = open(where, "w+")
        else:
            f = where
        try:
            f.write(self.marshal(node))
        finally:
            f.close()



# nxpy.xml package -----------------------------------------------------------

# Copyright Nicola Musatti 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
XML utility classes and functions.

Requires at least Python 2.6. Simple import breaks on Python 2.5

"""
from __future__ import absolute_import

import collections
import re

import lxml.etree

import six

import nxpy.core.error
import nxpy.core.past

nxpy.core.past.enforce_at_least(nxpy.core.past.V_2_6)


def parse(src):
    r"""parse the 'src' XML file and return a DOM."""
    
    parser = lxml.etree.XMLParser(strip_cdata=False)
    tree = lxml.etree.parse(src, parser)
    return tree


def make_property(elem, key=None):
    r"""
    Create a property on the text of element 'elem' or, if the 'key' argument is given, on its
    'key' attribute.
    
    """
    if key:
        def _get(self):
            return getattr(self, elem).get(key)
        def _set(self, value):
            getattr(self, elem).set(key, value)
            self._modified = True
        return property(_get, _set)
    else:
        def _get(self):
            return getattr(self, elem).text
        def _set(self, value):
            getattr(self, elem).text = value
            self._modified = True
        return property(_get, _set)


class QName(object):
    r"""Represent a qualified name"""

    _re = re.compile(r"\{(.*)\}(.*)")
    
    def __init__(self, tag):
        m = QName._re.match(tag)
        self.url = m.group(1)
        self.tag = m.group(2)

    @property
    def text(self):
        t = []
        if len(self.url) != 0:
            t.append("{{{0}}}".format(self.url))
        t.append(self.tag)
        return "".join(t)

    def __str__(self):
        return self.text()


class Namespace(object):
    r"""
    Represent an XML namespace and provide several utility functions that help handle a
    document without namespace tags.
    
    """
    def __init__(self, url="", element=None):
        if len(url) > 0 and element is not None:
            raise nxpy.core.error.ArgumentError(
                    "Only one between url and element should be specified")
        if element is not None:
            url = QName(element.tag).url
        self.url = url
        self.nspace = "{" + url + "}" if len(url) != 0 else ""

    def find(self, element, tag):
        return element.find(self.nspace + tag)

    def findall(self, element, tag):
        return element.findall(self.nspace + tag)

    def findtext(self, element, tag, default=None):
        return element.findtext(self.nspace + tag, default)

    def get_tag(self, element):
        return element.tag[len(self.nspace):]

    def Element(self, tag, attrib={}, **extra):
        return lxml.etree.Element(self.nspace + tag, attrib, **extra)
    
    def SubElement(self, parent, tag, attrib={}, **extra):
        return lxml.etree.SubElement(parent, self.nspace + tag, attrib, **extra)


class ContainerElementMixin(Namespace):
    r"""
    Mixin class for container workalikes backed by a DOM.
    
    """
    def __init__(self, parent, root_tag, namespace=""):
        super(ContainerElementMixin, self).__init__(namespace)
        self.parent = parent
        self.root_tag = root_tag
        self.root = self.find(self.parent, self.root_tag)
        self.modified = False

    def __len__(self):
        return len(self.root) if self.root is not None else 0


class MappingElementIterator(collections.Iterator):
    r"""
    Iterator over a 'MappingElement'.

    """
    def __init__(self, element):
        self.element = element
        self.iter = element.getchildren().iter()

    def __next__(self):
        return self.element.get_tag(next(self.iter))


class MappingElement(ContainerElementMixin, collections.MutableMapping):
    r"""
    Provide a tag/text map for the children of the 'root_tag' descendent of 'parent'. Given:

    .. code-block:: xml

        <parent>
            <root_tag>
                <key1>value1</key1>
                <key2>value2</key2>
            </root_tag>
        </parent>

    One could write::

        mappingElement["key1"] == "value1"
    
    """
    def __init__(self, parent, root_tag, namespace=""):
        ContainerElementMixin.__init__(self, parent, root_tag, namespace)

    def __getitem__(self, key):
        if self.root is None:
            raise KeyError()
        elem = self.find(self.root, key)
        if elem is None:
            raise KeyError()
        return elem.text

    def __setitem__(self, key, value):
        if self.root is None:
            self.root = self.SubElement(self.parent, self.root_tag)
        elem = self.find(self.root, key)
        if elem is None:
            elem = self.SubElement(self.root, key)
        self.modified = True
        elem.text = value

    def __delitem__(self, key):
        if self.root is None:
            raise KeyError()
        elem = self.find(self.root, key)
        if elem is None:
            raise KeyError()
        self.modified = True
        self.root.remove(elem)

    def __iter__(self):
        return MappingElementIterator(self)


class SequenceElement(ContainerElementMixin, collections.MutableSequence):
    r"""
    Provide a sequence of the values of the children tagged 'element_tag' of the 'root_tag' 
    descendent of 'parent'. Given:

    .. code-block:: xml

        <parent>
            <root_tag>
                <tag>value1</tag>
                <tag>value2</tag>
            </root_tag>
        </parent>

    One could write::

        sequenceElement[1] == "value2"
    
    """
    def __init__(self, parent, root_tag, element_tag, namespace="", indent="    "):
        ContainerElementMixin.__init__(self, parent, root_tag, namespace)
        self.element_tag = element_tag
        self.indent = indent
        self.elements = []
        i = 0
        if self.root is not None:
            for e in self.root:
                if e.tag is not lxml.etree.Comment and self.get_tag(e) == self.element_tag:
                    self.elements.append([i, e])
                i += 1

    def __len__(self):
        return len(self.elements) if self.elements is not None else 0

    def __getitem__(self, index):
        if self.root is None:
            raise IndexError()
        return self.elements[index][1].text

    def __setitem__(self, index, value):
        if self.root is None:
            if index != 0:
                raise IndexError()
            self.root = self.SubElement(self.parent, self.root_tag)
        elif index > len(self.root):
            raise IndexError()
        elem = None
        try:
            elem = self.elements[index][1]
        except IndexError:
            elem = self.SubElement(self.root, self.element_tag)
            self.elements.append([len(self.root) - 1, elem])
        elem.text = value
        self.modified = True

    def __delitem__(self, index):
        if self.root is None:
            raise IndexError()
        del self.root[self.elements[index][0]]
        del self.elements[index]
        for i in range(index, len(self.elements)):
            self.elements[i][0] -= 1
        self.modified = True

    def insert(self, index, value):
        if index > len(self.elements):
            raise IndexError()
        if self.root is None:
            self.root = self.SubElement(self.parent, self.root_tag)
        elem = self.Element(self.element_tag)
        elem.text = value
        elem.tail = self.root.tail + self.indent
        if index == len(self.elements):
            self.root.append(elem)
            self.elements.append([len(self.root) - 1, elem])
        else:
            self.root.insert(self.elements[index][0], elem)
            self.elements.insert(index, [self.elements[index][0], elem])
            for i in range(index + 1, len(self.elements)):
                self.elements[i][0] -= 1
        self.modified = True


class Writer(object):
    r"""
    Pretty-print an XML tree.
 
    """
    _name_re = re.compile(r"<([^\s]+)")
    _tag_re = re.compile(r"(</?)[^:]+:((:?[^>]+>)|(:?[^/]+/>))")
    _empty_re = re.compile(r"<([^/^>^ ]*)\s*/>")

    def __init__(self, root_tag, attributes=None, tab_size=0):
        self.root_tag = root_tag
        self.tab_size = tab_size
        self.attributes = attributes
        self.name = self._name_re.search(self.root_tag).group(1)
        self._root_re = re.compile(r"(<" + self.name + r"[^>]+>)")

    def marshal(self, node):
        s = None
        if nxpy.core.past.V_2_7.at_most():
                s = lxml.etree.tostring(node)
        else:
                s = lxml.etree.tostring(node, encoding="unicode")
        s = self._root_re.sub(self.root_tag, s, 1)
        s = self._empty_re.sub(r"<\1 />", s)
        if self.tab_size > 0:
            s = s.replace("\t", " " * self.tab_size)
        if self.attributes is not None:
            d = ( '<?xml version="' + self.attributes.get("version", "1.0") + 
                    '" encoding="' + self.attributes.get("encoding", "UTF-8") + '"')
            if "standalone" in self.attributes:
                d += ' standalone="' + self.attributes["standalone"] + '"'
            d += "?>\n"
            s = d + s
        return s + "\n\n"

    def write(self, node, where):
        f = open(where, "w+") if isinstance(where, six.string_types) else where
        try:
            f.write(self.marshal(node))
        finally:
            f.close()



# nxpy.wx package ------------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Interactive program driver - wxPython version.

An alternative implementation of *command.interpreter.Interpreter* for use with wxPython 
applications, as the default one is not compatible with wxPython's application loop.

"""

from __future__ import absolute_import

import wx

import nxpy.command.interpreter


class Popen(object):
    r"""A simple wxPython based, Popen-like object."""
    def __init__(self, cmd):
        self.process = wx.Process(None)
        self.process.Redirect();
        wx.Execute(cmd, wx.EXEC_ASYNC, self.process)

    def send(self, data):
        self.process.GetOutputStream().write(data)

    def recv(self):
        stream = self.process.GetInputStream()
        if stream.CanRead():
            return stream.read()
        return ""

    def recv_err(self):
        stream = self.process.GetErrorStream()
        if stream.CanRead():
            return stream.read()
        return ""


class Interpreter(nxpy.command.interpreter.BaseInterpreter):
    r"""A command interpreter suitable for use in wxPython based applications."""
    def __init__(self, cmd):
        super(Interpreter, self).__init__(Popen(cmd))



# nxpy.msvs package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Manipulation of Microsoft Visual Studio project files.

Currently supports Visual Studio 2008 (9.0) only.

Requires at least Python 2.6

"""

from __future__ import absolute_import

import os.path
import re

import nxpy.xml.util
import nxpy.msvs.assembly_info


_ns = nxpy.xml.util.Namespace("http://schemas.microsoft.com/developer/msbuild/2003")


class BadProjectFileError(Exception):
    pass


class MissingConfigurationError(Exception):
    pass


class UnknownOutputTypeError(Exception):
    pass


class Reference(object):
    def __init__(self, element):
        self.element = element
        name = element.get("Include")
        i = name.find(",")
        if i != -1:
            self.name = name[0:i]
        else:
            self.name = name
    
    @property
    def hint_path(self):
        return _ns.findtext(self.element, "HintPath")

    def __str__(self):
        s = self.name
        if self.hint_path:
            s = s + "\t" + self.hint_path
        return s


class Satellite(object):
    def __init__(self, name, lang, build_path):
        self.name = name
        self.lang = lang
        self.build_path = build_path
        
    @property
    def target(self):
        return os.path.join(self.build_path, self.lang, self.name)
    
    def __str__(self):
        return self.target


class Project(object):
    _app_conf_re = re.compile(r"[^.]+\.config", re.IGNORECASE)
    
    _emb_res_re = re.compile(r"[^.]+\.([^.]+)\.resx", re.IGNORECASE)
    
    def __init__(self, path):
        self.path = os.path.realpath(path)
        self.dir, name = os.path.split(self.path)
        self.name, self.extension = os.path.splitext(name)
        try:
            self.tree = nxpy.xml.util.parse(path)
        except IOError:
            raise BadProjectFileError(path)
        self.root = self.tree.getroot()
        pgs = [ pg for pg in _ns.findall(self.root, "PropertyGroup") ]
        self.configuration = _ns.findtext(pgs[0], "Configuration")
        self.platform = _ns.findtext(pgs[0], "Platform")
        self.output_type = _ns.findtext(pgs[0], "OutputType")
        self.assembly_name = _ns.findtext(pgs[0], "AssemblyName")
        self.assembly_file = self.assembly_name + self._extension()
        for pg in pgs[1:]:
            cond = pg.get("Condition")
            if ( cond and ( cond.find("Configuration") == -1 or 
                            cond.find(self.configuration) != -1 ) and
                          ( cond.find("Platform") == -1 or 
                            cond.find(self.platform) != -1 ) ):
                self.output_path = _ns.findtext(pg, "OutputPath")
                self.assembly_dir = os.path.join(self.dir, self.output_path)
                break
        else:
            raise MissingConfigurationError()
        self.assembly = os.path.join(self.assembly_dir, self.assembly_file)
        self.references = []
        self.sources = []
        self.satellites = []
        self.app_config = None
        self.assembly_info = None
        for ig in _ns.findall(self.root, "ItemGroup"):
            for ref in _ns.findall(ig, "Reference"):
                self.references.append(Reference(ref))
            for tag in ( "ApplicationDefinition", "Page", "Resource" ):
                for src in _ns.findall(ig, tag):
                    self._add_source(src)
            for src in _ns.findall(ig, "Compile"):
                self._add_source(src)
                if src.get("Include").find("AssemblyInfo") != -1:
                    self.assembly_info = nxpy.msvs.assembly_info.AssemblyInfo(
                            os.path.join(self.dir, *src.get("Include").split("\\")))
            for res in _ns.findall(ig, "EmbeddedResource"):
                self._add_source(res)
                match = self._emb_res_re.match(res.get("Include"))
                if match and match.group(1) not in [ l.lang for l in self.satellites ]:
                    self.satellites.append(Satellite(self.assembly_name + ".resources.dll", 
                            match.group(1), self.assembly_dir))
            for f in _ns.findall(ig, "None"):
                match = self._app_conf_re.match(f.get("Include"))
                if match:
                    self._add_source(f)
                    self.app_config = os.path.join(self.assembly_dir, self.assembly_file + 
                            ".config")

    def _add_source(self, src):                
        self.sources.append(os.path.join(self.dir, src.get("Include")))

    def _extension(self):
        t = self.output_type.lower()
        if t == "library":
            return ".dll"
        elif t == "winexe":
            return ".exe"
        else:
            raise UnknownOutputTypeError(self.output_type)
            
    def __str__(self):
        s = ( "Project: " + self.path + "\nConfiguration: " + self.configuration + "\nPlatform: " + 
              self.platform + "\nAssembly: " + self.assembly + "\nVersion: " + 
              self.assembly_info.version + "\nApp.config: " + str(self.app_config) + 
              "\nReferences:" )
        return ( "\n\t".join([ s ] + [ str(r) for r in self.references ]) + 
                 "\n\t".join([ "\nSatellites:" ] + 
                         [ str(sat) for sat in self.satellites ]) + 
                 "\n\t".join([ "\nSources:" ] + [ str(src) for src in self.sources ]) )



# nxpy.scons package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2012
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Support for handling project files with SCons, i.e. use it as a meta-build tool.

"""

class Project(object):
    def __init__(self):
        self.projects = {}
    
    def add(self, env, target, path):
        self.projects[str(target)] = path
        
    def __call__(self, target, source, env, for_signature):
        return self.projects[str(target)]


# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
File related utilities.

"""

# Python 2.5 compatibility
from __future__ import with_statement    
from __future__ import absolute_import

import os

import six
from six.moves import zip

import nxpy.core.past


def compare(file1, file2, ignore_eof=True, encoding=None):
    r"""
    Compare two text files for equality.
    If *ignore_eof* is *True*, end of line characters are not considered.
    If not *None* *encoding* is used to open the files. On Python 2.x *encoding* is ignored.
    
    """
    if isinstance(file1, six.string_types):
        f1 = open_(file1, "r", encoding=encoding)
    else:
        f1 = file1
        f1.seek(0, os.SEEK_SET)
    if isinstance(file2, six.string_types):
        f2 = open_(file2, "r", encoding=encoding)
    else:
        f2 = file2
        f2.seek(0, os.SEEK_SET)
    with f1:
        with f2:
            for l1, l2 in zip(f1, f2):
                if ( ( not ignore_eof and l1 != l2 ) or 
                     ( ignore_eof and l1.rstrip("\r\n") != l2.rstrip("\r\n") ) ):
                    return False
            return True


def open_(*args, **kwargs):
    r"""Open a file removing invalid arguments on Python 2.x."""
    if nxpy.core.past.V_2_7.at_most() and "encoding" in kwargs:
        del kwargs["encoding"]
    return open(*args, **kwargs)



# nxpy.msvs._impl.solution package -------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

from __future__ import absolute_import

import re


class Description(object):
    descRe = re.compile('Description[ \t]*=(.*)')
    
    def __init__(self, value):
        self.len = len(value[0])
        self.pos = value[1] - self.len
        m = self.descRe.match(value[0])
        if m and m.groups is not None:
            self.value = m.group(1).strip()
        else:
            self.value = ""


class Project(object):
    def __init__(self, name, path):
        self.name = name
        self.path = path

    def __str__(self):
        return " ".join([ "Project:", self.name, "-", self.path ]) 


class Solution(object):
    def __init__(self, text, version, projects, description):
        self.text = text
        self.version = version
        self.projects = projects
        self.description = description
        
    def __str__(self):
        return "\n\t".join([ "Solution " + self.version, ] + [ str(p) for p in self.projects ])



# nxpy.maven package ---------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Maven wrapper.

"""

from __future__ import absolute_import

import logging

import nxpy.command.command
import nxpy.command.option
import nxpy.core.sequence

import nxpy.core._impl.log

_log = logging.getLogger(__name__)

_config = nxpy.command.option.Config(
        separator = ",",
        iterable_opts = ( "projects" ), )


class Mvn(nxpy.command.command.Command):
    def __init__(self, debug=None):
        super(Mvn, self).__init__("mvn", debug)

    def _make_options(self, projects):
        kwargs = {}
        if projects is not None:
            kwargs["projects"] = nxpy.core.sequence.make_tuple(projects)
        return kwargs
    
    def clean(self, projects=None, debug=None):
        op = nxpy.command.option.Parser(_config, None, ( "clean", ), {},
                **self._make_options(projects))
        self.run(op, debug)

    def deploy(self, projects=None, debug=None):
        op = nxpy.command.option.Parser(_config, None, ( "deploy", ), {},
                **self._make_options(projects))
        self.run(op, debug)

    def package(self, projects=None, debug=None):
        op = nxpy.command.option.Parser(_config, None, ( "package", ), {},
                **self._make_options(projects))
        self.run(op, debug)



# nxpy package ---------------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2017
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Packaging information.

"""

from __future__ import absolute_import

import codecs
import os.path

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here,'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name = 'Nxpy',
    version = '0.5.0',
    url = 'http://nxpy.sourceforge.net',
    author = 'Nicola Musatti',
    author_email = 'nmusatti@users.sf.net',
    packages = find_packages(exclude=["*._test", "*._test.*", "_test.*", "_test"]),
    #package_data = { '' : [ '*.txt', '*.rst', ], },
    install_requires = ['six >= 1.8.0'],
    license = 'Boost Software License version 1.0',
    description = "Nick's Python Toolchest",
    long_description = long_description,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',

    ],
)



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Memoize class instances according to a given key.
 
By default the key only assumes the *True* value, thus implementing a singleton. 

"""

from __future__ import absolute_import

import nxpy.core.abstract
import nxpy.core.past

if nxpy.core.past.V_2_6.at_least():
    import abc
    import six


class Memo(object):
    r"""
    Base class for classes that require memoization.
    
    Subclasses should override the *_key(\*args, \*\*kwargs)* method to compute a key on the
    constructor's arguments.
    
    Care should be taken to avoid calling *__init__()* again for entities already constructed.
    
    """
    #if nxpy.core.past.V_2_6.at_least():
    #    __metaclass__ = abc.ABCMeta
    
    _dict = {}
    
    def __new__(cls, *args, **kwargs):
        r"""
        Return the instance corresponding to the given key, creating it if it doesn't exist.
        
        """
        k = cls._key(*args, **kwargs)
        try:
            return Memo._dict[(cls, k)]
        except KeyError:
            i = super(Memo, cls).__new__(cls)
            Memo._dict[(cls, k)] = i
            return i

    if nxpy.core.past.V_2_6.at_least():
        @nxpy.core.abstract.abstractstatic
        def _key(*args, **kwargs):
            r"""
            Returns the key by which instances are memoized.
    
            Subclasses should override it as a function of their construction arguments.
            
            """
            return True
    else:
        @staticmethod
        def _key(*args, **kwargs):
            return True


if nxpy.core.past.V_2_6.at_least():
    Memo = six.add_metaclass(abc.ABCMeta)(Memo)



# nxpy.svn package -----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Subversion administration tool wrapper.

"""

from __future__ import absolute_import

import os

import nxpy.command.command
import nxpy.command.option


_config = nxpy.command.option.Config()

class SvnAdmin(nxpy.command.command.Command):
    def __init__(self, debug=None):
        super(SvnAdmin, self).__init__("svnadmin", debug)
    
    def create(self, path, debug=None):
        op = nxpy.command.option.Parser(_config, "create", ( path, ), {})
        self.run(op, debug)
        return "file:///" + path.replace(os.sep, "/").lstrip("/")



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2008 - 2015
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Helper classes for the implementation of read-only and writable file objects that forward calls
to an actual file object variable.

"""

from __future__ import absolute_import


class ReadOnlyFileObject(object):
    r"""
    Implement the non modifying portion of the file object protocol by delegating to 
    another file object.
    
    Subclass and override as needed.
    """
    def __init__(self, file_=None):
        r"""Set the delegate file object."""
        self.setFile(file_)

    def setFile(self, file_):
        r"""Set the delegate file object."""
        self._file = file_

    def __iter__(self):
        return self._file.__iter__()

    def __next__(self):
        return self._file.__next__()

    def __getattr__(self, name):
        if name in ( "truncate", "write", "writeline" ):
            raise AttributeError()
        return getattr(self._file, name)


class WritableFileObject(ReadOnlyFileObject):
    r"""
    Implement the file object protocol by delegating to another file object.
    
    Subclass and override as needed.
    """
    def __init__(self, file_=None):
        super(WritableFileObject, self).__init__(file_)

    def __getattr__(self, name):
        return getattr(self._file, name)



# nxpy.core package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Sort functions.

"""

def topological_sort(pairs):
    r"""
    Provide a topological ordering of the supplied pair elements.
    
    *pairs* is a sequence of two element sequences, in which the first element comes before
    the second according to the desired ordering criterium.
    
    """
    res = []
    top = set()
    nodes = {}
    for end, start in pairs:
        if end not in nodes:
            nodes[end] = []
        nodes[end].append(start)
    for end, start in pairs:
        if start not in nodes:
            top.add(start)
    while len(top) > 0:
        s = top.pop()
        res.append(s)
        for e in nodes:
            try:
                nodes[e].remove(s)
                if len(nodes[e]) == 0:
                    top.add(e)
            except ValueError:
                pass
        for e in top:
            try:
                del nodes[e]
            except:
                pass
    if len(nodes) > 0:
        return None
    else:
        return res



# nxpy.test package ----------------------------------------------------------

# Copyright Nicola Musatti 2011 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Logging configuration for tests.

"""

from __future__ import absolute_import

import logging
import os
import sys

_level = logging.INFO

try:
    _level = logging.__dict__[os.environ["NXPY_TEST_LOG_LEVEL"]]
except KeyError:
    pass

logging.basicConfig(stream=sys.__stderr__, level=_level)



# nxpy.command package -------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Exception classes for the nxpy.command package.

"""


class Error(Exception):
    r"""Package exceptions' base class."""


class BadLogFormat(Error):
    r"""Raised if the requested formatting option is unknown."""


class ExpectError(Exception):
    """Raised on invalid input from stdout or stderr."""


class TimeoutError(Exception):
    """Raised when expect didn't satisfy a timing constraint."""


class TimerError(Exception):
    """Raised on misuse of the Timer class."""



# nxpy.test package ----------------------------------------------------------

# Copyright Nicola Musatti 2010 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Environment configuration for tests that interact with the system.

"""

from __future__ import absolute_import

import os
import os.path

import nxpy.core.memo


class TestEnvNotSetError(Exception):
    r"""Raised when the test environment hasn't been setup, i.e. NXPY_TEST_DIR is not set."""


class _Dir(nxpy.core.memo.Memo):
    def __init__(self):
        if getattr(self, "base", None) is None:
            try:
                self.base = os.environ["NXPY_TEST_DIR"]
                if not os.path.isdir(self.base):
                    self.base = None
            except KeyError:
                self.base = None

    @staticmethod
    def _key(*args, **kwargs):
        return super(_Dir, _Dir)._key(*args, **kwargs)


class EnvBase(object):
    def __init__(self, elem):
        self._dir = _Dir()
        if self._dir.base is None:
            raise TestEnvNotSetError()
        self.elem = elem
        self.elem_dir = os.path.join(self._dir.base, elem)

    
class Data(EnvBase):
    def __init__(self, package):
        EnvBase.__init__(self, "data")
        self.data = os.path.join(self.elem_dir, package)


def get_data(test, package):
    try:
        return Data(package)
    except TestEnvNotSetError:
        test.skipTest("Test environment not set")
        return None


class Env(EnvBase):
    def __init__(self, package):
        EnvBase.__init__(self, "wcopy")
        self.wcopy = os.path.join(self.elem_dir, package)
        rd = self._dir.base
        repo_dir = []
        while True:
            rd, d = os.path.split(rd)
            if d:
                repo_dir.append(d)
            else:
                rd = rd.replace(os.path.sep, "")
                if rd:
                    repo_dir.append(rd)
                break
        repo_dir.reverse()
        repo_dir.extend(("repo", package))
        self.repo = "file:///" + "/".join(repo_dir)
        self.backup = os.path.join(self._dir.base, "backup", package)


def get_env(test, package):
    try:
        return Env(package)
    except TestEnvNotSetError:
        test.skipTest("Test environment not set")
        return None



# nxpy.command package -------------------------------------------------------

# Copyright Nicola Musatti 2011 - 2014
# Use, modification, and distribution are subject to the Boost Software
# License, Version 1.0. (See accompanying file LICENSE.txt or copy at
# http://www.boost.org/LICENSE_1_0.txt)

# See http://nxpy.sourceforge.net for library home page. ---------------------

r"""
Tools to wrap a Python API around interactive and non-interactive programs.

The :py:class:`.command.Command` and :py:class:`interpreter.Interpreter` classes handle batch and 
interactive commands respectively. They can be provided with :py:class:`.option.Config` instances
which describe the options available to the programs being wrapped. The :py:class:`.option.Parser` 
class can then be used to validate option sets and construct the corresponding command lines.
See the :py:mod:`.svn.svn` module for a concrete example.

"""
